from __future__ import annotations

from abc import ABC
from typing import Annotated, ClassVar, Literal, Self, Type, TypeAlias, Union

from pydantic import BaseModel, Field, model_validator
from pydantic.config import ConfigDict
from pydantic.fields import FieldInfo
from typing_extensions import Unpack

from bayesline.api._src.equity.riskmodels_settings import FactorRiskModelSettings
from bayesline.api._src.registry import (
    EmptySettingsMenu,
    Settings,
    SettingsMenu,
    SettingsTypeMetaData,
)


class ReportAccessorSettings(Settings[EmptySettingsMenu]):
    """Settings for report accessor configuration.

    This class defines the configuration for accessing and processing reports,
    including axes, metrics, pivot columns, and additional report linking.
    """

    axes: dict[str, list[str]]
    metric_cols: list[str]
    pivot_cols: list[str]

    # can be used for multi accessors, one entry for each underlying report
    # for example: extra_cols = name, risk_model, date
    # for example: extra_paths = [["My Setting", "US-All", "2024-12-12"]]
    extra_paths: list[list[str]] = Field(default_factory=list)
    extra_cols: list[str] = Field(default_factory=list)

    linked_reports: list[int] = Field(default_factory=list)
    tags: list[str] = Field(default_factory=list)

    @property
    def menu_type(self) -> type[EmptySettingsMenu]:  # noqa: D102
        return EmptySettingsMenu

    def describe(self, menu: EmptySettingsMenu) -> str:
        """Describe the report accessor settings.

        Parameters
        ----------
        menu : EmptySettingsMenu
            The menu to use to describe the settings.

        Returns
        -------
        str
            A description of the report accessor settings.
        """
        return f"Report accessor with {len(self.axes)} axes and {len(self.metric_cols)} metrics"


class MeasureSettings(BaseModel, ABC, frozen=True, extra="forbid"):
    """Defines settings for a measure."""

    type: str

    @property
    def columns(self) -> list[str]:
        """Get the column names for this measure.

        Returns
        -------
        list[str]
            The column names for this measure.
        """
        return [self.type]

    def __init_subclass__(cls, **kwargs: Unpack[ConfigDict]) -> None:
        """Initialize subclass and validate that all fields have default values.

        This method ensures that every field in the subclass has a default value,
        so settings can be instantiated without any arguments.

        Parameters
        ----------
        **kwargs : Unpack[ConfigDict]
            Configuration dictionary arguments.

        Raises
        ------
        AssertionError
            If any field does not have a default value.
        """
        super().__init_subclass__(**kwargs)
        # check that every field has a default value, so we can instantiate settings
        # without any argument. Fields defined as a: int = 1, a: int = Field(1), and
        # a: list[int] = Field(default_factory=list) are all valid.
        for f in cls.__annotations__:
            msg = f"{f} not does not have a default value"
            if f not in cls.__dict__:
                raise AssertionError(msg)  # fields without ... = Field(...)
            if isinstance(cls.__dict__[f], FieldInfo):
                if cls.__dict__[f].is_required():
                    raise AssertionError(msg)  # with = Field(...)


class PassThroughFactor2DMeasureSettings(MeasureSettings):
    """Settings for pass-through factor 2D measure.

    This class defines settings for a pass-through factor 2D measure,
    which allows passing through factor data in 2D format.
    """

    type: Literal["PassThroughFactor2D"] = Field("PassThroughFactor2D", repr=False)
    name: str = Field("PassThroughFactor2D", description="The name of the measure.")

    @property
    def columns(self) -> list[str]:
        """Get the column names for this measure.

        Returns
        -------
        list[str]
            The column names for this measure.
        """
        return [self.name]


class HoldingsMeasureSettings(MeasureSettings):
    """Settings for holdings measure.

    This class defines settings for a holdings measure,
    which provides portfolio, benchmark, and active holdings data.
    """

    type: Literal["Holdings"] = Field("Holdings", repr=False)

    @property
    def columns(self) -> list[str]:
        """Get the column names for this measure.

        Returns
        -------
        list[str]
            The column names for this measure.
        """
        return ["Portfolio", "Benchmark", "Active"]


class ExposureMeasureSettings(MeasureSettings):
    """Settings for exposure measure.

    This class defines settings for an exposure measure,
    which provides portfolio, benchmark, and active exposure data.
    """

    type: Literal["Exposure"] = Field("Exposure", repr=False)
    rescale_bench: bool = Field(
        True,
        description="Rescale the benchmark holdings to sum to the sum of the holdings.",
    )
    normalize_holdings: bool = Field(True, description="Make holdings sum to one.")

    @property
    def columns(self) -> list[str]:
        """Get the column names for this measure.

        Returns
        -------
        list[str]
            The column names for this measure.
        """
        return ["Portfolio", "Benchmark", "Active"]


class XSigmaRhoMeasureSettings(MeasureSettings):
    """Settings for X-Sigma-Rho measure.

    This class defines settings for an X-Sigma-Rho measure,
    which provides exposure, volatility, correlation, and contribution data.
    """

    type: Literal["XSigmaRho"] = Field("XSigmaRho", repr=False)
    rescale_bench: bool = Field(
        True,
        description="Rescale the benchmark holdings to sum to the sum of the holdings.",
    )
    normalize_holdings: bool = Field(True, description="Make holdings sum to one.")
    analytics_space: Literal["absolute", "active", "benchmark"] = Field(
        "absolute",
        description=(
            "Compute the analytics in `absolute` space, `active` space, or compute the "
            "analytics only on the `benchmark`."
        ),
    )

    @property
    def columns(self) -> list[str]:
        """Get the column names for this measure.

        Returns
        -------
        list[str]
            The column names for this measure.
        """
        return [
            "Exposure (X)",  # X
            "Volatility (\u03c3)",  # σ
            "Correlation (\u03c1)",  # ρ
            "Contribution (X\u03c3\u03c1)",  # Xσρ
        ]


class TimeSeriesXSigmaRhoMeasureSettings(MeasureSettings):
    """Settings for time series X-Sigma-Rho measure.

    This class defines settings for a time series X-Sigma-Rho measure,
    which provides time series data for exposure, volatility, correlation, and contribution.
    """

    type: Literal["TimeSeriesXSigmaRho"] = Field("TimeSeriesXSigmaRho", repr=False)
    rescale_bench: bool = Field(
        True,
        description="Rescale the benchmark holdings to sum to the sum of the holdings.",
    )
    normalize_holdings: bool = Field(True, description="Make holdings sum to one.")
    analytics_space: Literal["absolute", "active", "benchmark"] = Field(
        "absolute",
        description=(
            "Compute the anlytics in `absolute` space, `active` space, or compute the "
            "analytics only on the `benchmark`."
        ),
    )
    backfill_holdings: bool = Field(
        False,
        description="Backfill the latest holdings, accounting for listing/delisting.",
    )

    @property
    def columns(self) -> list[str]:
        """Get the column names for this measure.

        Returns
        -------
        list[str]
            The column names for this measure.
        """
        return [
            "Exposure (X)",  # X
            "Volatility (\u03c3)",  # σ
            "Correlation (\u03c1)",  # ρ
            "Contribution (X\u03c3\u03c1)",  # Xσρ
        ]


class RiskDecompositionMeasureSettings(MeasureSettings):
    """Settings for risk decomposition measure.

    This class defines settings for a risk decomposition measure. Currently exposure,
    stand-alone volatility and variance contribution are computed.
    """

    type: Literal["RiskDecomposition"] = Field("RiskDecomposition", repr=False)
    rescale_abs_bench: bool = Field(
        True,
        description=(
            "Rescale the benchmark holdings to absolute sum to the absolute sum of the "
            "holdings."
        ),
    )
    analytics_space: Literal["absolute", "active", "benchmark"] = Field(
        "absolute",
        description=(
            "Compute the anlytics in `absolute` space, `active` space, or compute the "
            "analytics only on the `benchmark`."
        ),
    )

    @property
    def columns(self) -> list[str]:
        """Get the column names for this measure.

        Returns
        -------
        list[str]
            The column names for this measure.
        """
        return [
            "Exposure",
            "StandaloneVolatility",
            "VarianceContribution",
        ]


class TimeSeriesBetaMeasureSettings(MeasureSettings):
    """Settings for time series beta measure.

    This class defines settings for a time series beta measure,
    which provides time series beta data for portfolio analysis.
    """

    type: Literal["TimeSeriesBeta"] = Field("TimeSeriesBeta", repr=False)
    rescale_bench: bool = Field(
        True,
        description="Rescale the benchmark holdings to sum to the sum of the holdings.",
    )

    @property
    def columns(self) -> list[str]:
        """Get the column names for this measure.

        Returns
        -------
        list[str]
            The column names for this measure.
        """
        return [
            "Beta (\u03b2)",  # β
        ]


class HVaRMeasureSettings(MeasureSettings):
    """Settings for historical VaR measure.

    This class defines settings for a historical Value at Risk (VaR) measure,
    which provides historical VaR calculations at specified confidence levels.
    """

    type: Literal["Historical VaR"] = Field("Historical VaR", repr=False)
    alpha: list[float] = Field([0.1, 0.05])

    @property
    def columns(self) -> list[str]:
        """Get the column names for this measure.

        Returns
        -------
        list[str]
            The column names for this measure.
        """
        return [f"HistVaR {a:.0%}" for a in self.alpha]


class AVaRMeasureSettings(MeasureSettings):
    """Settings for analytical VaR measure.

    This class defines settings for an analytical Value at Risk (VaR) measure,
    which provides analytical VaR calculations at specified confidence levels.
    """

    type: Literal["Analytical VaR"] = Field("Analytical VaR", repr=False)
    alpha: list[float] = Field([0.1, 0.05])

    @property
    def columns(self) -> list[str]:
        """Get the column names for this measure.

        Returns
        -------
        list[str]
            The column names for this measure.
        """
        return [f"AnVaR {a:.0%}" for a in self.alpha]


class TimeSeriesVolatilityMeasureSettings(MeasureSettings):
    """Settings for time series volatility measure.

    This class defines settings for a time series volatility measure,
    which provides time series volatility data using rolling windows.
    """

    type: Literal["TimeSeriesVolatility"] = Field("TimeSeriesVolatility", repr=False)
    window: str = Field("6mo", description="The size of the rolling window")
    ddof: int = Field(1, description="The degrees of freedom to use for the volality")


class CumsumMeasureSettings(MeasureSettings):
    """Settings for cumulative sum measure.

    This class defines settings for a cumulative sum measure,
    which provides cumulative sum calculations for time series data.
    """

    type: Literal["Cumsum"] = Field("Cumsum", repr=False)


class DrawdownMeasureSettings(MeasureSettings):
    """Settings for drawdown measure.

    This class defines settings for a drawdown measure,
    which provides drawdown calculations for portfolio analysis.
    """

    type: Literal["Drawdown"] = Field("Drawdown", repr=False)
    window: str = Field("6mo", description="The size of the rolling window")


class MovingAverageMeasureSettings(MeasureSettings):
    """Settings for moving average measure.

    This class defines settings for a moving average measure,
    which provides moving average calculations for time series data.
    """

    type: Literal["MovingAverage"] = Field("MovingAverage", repr=False)
    window: str = Field("6mo", description="The size of the rolling window")
    var: Literal["r2", "sigma2", "sigma", "aic", "bic"] = Field(
        "r2", description="The variable to average"
    )

    @property
    def columns(self) -> list[str]:
        """Get the column names for this measure.

        Returns
        -------
        list[str]
            The column names for this measure.
        """
        return [f"{self.type} {self.var}"]


class FactorMovingAverageMeasureSettings(MeasureSettings):
    """Settings for factor moving average measure.

    This class defines settings for a factor moving average measure,
    which provides moving average calculations for factor data.
    """

    type: Literal["FactorMovingAverage"] = Field("FactorMovingAverage", repr=False)
    window: str = Field("6mo", description="The size of the rolling window")
    var: Literal["abs_t", "p", "p1", "p5"] = Field(
        "p", description="The variable to average"
    )

    @property
    def columns(self) -> list[str]:
        """Get the column names for this measure.

        Returns
        -------
        list[str]
            The column names for this measure.
        """
        return [f"{self.type} {self.var}"]


class MovingAverageRSquaredMeasureSettings(MeasureSettings):
    """Settings for moving average R-squared measure.

    This class defines settings for a moving average R-squared measure,
    which provides moving average R-squared calculations for time series data.
    """

    type: Literal["MovingAverageRSquared"] = Field("MovingAverageRSquared", repr=False)
    rescale_bench: bool = Field(
        True,
        description="Rescale the benchmark holdings to sum to the sum of the holdings.",
    )
    normalize_holdings: bool = Field(True, description="Make holdings sum to one.")
    window: str = Field("6mo", description="The size of the rolling window")


class VolForecastMeasureSettings(MeasureSettings):
    """Settings for volatility forecast measure.

    This class defines settings for a volatility forecast measure,
    which provides volatility forecasting calculations and metrics.
    """

    type: Literal["VolForecast"] = Field("VolForecast", repr=False)
    window: str = Field("6mo", description="The size of the rolling window")
    metric: Literal["qlike", "mse", "mae", "bias"] = Field(
        "qlike", description="The metric to use"
    )
    ddof: int = Field(1, description="The degrees of freedom to use for the stability")

    @property
    def columns(self) -> list[str]:
        """Get the column names for this measure.

        Returns
        -------
        list[str]
            The column names for this measure.
        """
        return [
            "VolForecast",
            "VolForecastRealized",
            "VolForecastStability",
            "VolForecastLoss",
        ]


class FactorVolForecastMeasureSettings(MeasureSettings):
    """Settings for factor volatility forecast measure.

    This class defines settings for a factor volatility forecast measure,
    which provides factor volatility forecasting calculations and metrics.
    """

    type: Literal["FactorVolForecast"] = Field("FactorVolForecast", repr=False)
    window: str = Field("6mo", description="The size of the rolling window")
    metric: Literal["qlike", "mse", "mae", "bias"] = Field(
        "qlike", description="The metric to use"
    )
    ddof: int = Field(1, description="The degrees of freedom to use for the stability")

    @property
    def columns(self) -> list[str]:
        """Get the column names for this measure.

        Returns
        -------
        list[str]
            The column names for this measure.
        """
        return [
            "VolForecast",
            "VolForecastRealized",
            "VolForecastStability",
            "VolForecastLoss",
        ]


class FactorCovarianceMeasureSettings(MeasureSettings):
    """Settings for factor covariance measure.

    This class defines settings for a factor covariance measure,
    which provides factor covariance calculations.
    """

    type: Literal["FactorCovariance"] = Field("FactorCovariance", repr=False)


class IdiosyncraticReturnMeasureSettings(MeasureSettings):
    """Settings for idiosyncratic return measure.

    This class defines settings for an idiosyncratic return measure,
    which provides idiosyncratic return calculations.
    """

    type: Literal["IdiosyncraticReturn"] = Field("IdiosyncraticReturn", repr=False)


class IdiosyncraticVolatilityMeasureSettings(MeasureSettings):
    """Settings for idiosyncratic volatility measure.

    This class defines settings for an idiosyncratic volatility measure,
    which provides idiosyncratic volatility calculations.
    """

    type: Literal["IdiosyncraticVolatility"] = Field(
        "IdiosyncraticVolatility", repr=False
    )


class FactorIdioMeasureSettings(MeasureSettings):
    """Settings for factor idiosyncratic measure.

    This class defines settings for a factor idiosyncratic measure,
    which provides factor idiosyncratic calculations.
    """

    type: Literal["FactorIdio"] = Field("FactorIdio", repr=False)


class ForecastBacktestMeasureSettings(MeasureSettings):
    """Settings for forecast backtest measure.

    This class defines settings for a forecast backtest measure,
    which provides forecast backtest calculations and metrics.
    """

    type: Literal["ForecastBacktest"] = Field("ForecastBacktest", repr=False)
    window: str = Field("6mo", description="The size of the rolling window")
    ddof: int = Field(1, description="The degrees of freedom to use for the stability")

    @property
    def columns(self) -> list[str]:
        """Get the column names for this measure.

        Returns
        -------
        list[str]
            The column names for this measure.
        """
        return ["GlobalMinVolPortVolatility", "GlobalMinVolPortPredictedVolatility"]


class BrinsonAttributionMeasureSettings(MeasureSettings):
    """Settings for Brinson attribution measure.

    This class defines settings for a Brinson attribution measure,
    which provides Brinson attribution calculations for portfolio analysis.
    """

    type: Literal["BrinsonAttribution"] = Field("BrinsonAttribution", repr=False)
    rescale_bench: bool = Field(
        True,
        description="Rescale the benchmark holdings to sum to the sum of the holdings.",
    )
    normalize_holdings: bool = Field(True, description="Make holdings sum to one.")
    multiperiod_aggregation: Literal["none", "optimized"] = Field("none")
    return_aggregation_type: Literal["arithmetic", "geometric"] = Field("geometric")

    @property
    def columns(self) -> list[str]:
        """Get the column names for this measure.

        Returns
        -------
        list[str]
            The column names for this measure.
        """
        return ["Allocation", "Selection", "Interaction", "Active"]


class FactorAttributionMeasureSettings(MeasureSettings):
    """Settings for factor attribution measure.

    This class defines settings for a factor attribution measure,
    which provides factor attribution calculations for portfolio analysis.
    """

    type: Literal["FactorAttribution"] = Field("FactorAttribution", repr=False)
    rescale_bench: bool = Field(
        True,
        description="Rescale the benchmark holdings to sum to the sum of the holdings.",
    )
    normalize_holdings: bool = Field(True, description="Make holdings sum to one.")
    multiperiod_aggregation: Literal["none", "optimized"] = Field("none")
    return_aggregation_type: Literal["arithmetic", "geometric"] = Field("geometric")

    @property
    def columns(self) -> list[str]:
        """Get the column names for this measure.

        Returns
        -------
        list[str]
            The column names for this measure.
        """
        return ["Portfolio", "Benchmark", "Active"]


class FactorAttributionDrilldownMeasureSettings(MeasureSettings):
    """Settings for factor attribution drilldown measure.

    This class defines settings for a factor attribution drilldown measure,
    which provides factor attribution drilldown calculations for portfolio analysis.
    """

    type: Literal["FactorAttributionDrilldown"] = Field(
        "FactorAttributionDrilldown", repr=False
    )
    rescale_bench: bool = Field(
        True,
        description="Rescale the benchmark holdings to sum to the sum of the holdings.",
    )
    normalize_holdings: bool = Field(True, description="Make holdings sum to one.")
    multiperiod_aggregation: Literal["none", "optimized"] = Field("none")
    return_aggregation_type: Literal["arithmetic", "geometric"] = Field("geometric")

    @property
    def columns(self) -> list[str]:
        """Get the column names for this measure.

        Returns
        -------
        list[str]
            The column names for this measure.
        """
        if self.normalize_holdings:
            return ["Return", "BenchmarkReturn", "ActiveReturn"]
        else:
            return ["PnL", "BenchmarkPnL", "ActivePnL"]

    @model_validator(mode="after")
    def validate_settings(self) -> Self:
        if not self.normalize_holdings and self.return_aggregation_type == "geometric":
            raise ValueError(
                "Geometric return aggregation is not supported when holdings are not "
                "normalized."
            )
        return self


class PortfolioStressTestMeasureSettings(MeasureSettings):
    """Settings for portfolio stress test measure.

    This class defines settings for a portfolio stress test measure,
    which provides portfolio stress test calculations.
    """

    type: Literal["PortfolioStressTest"] = Field("PortfolioStressTest", repr=False)
    shock_size: float = Field(0.1, description="The size of the shock to apply.")

    @property
    def columns(self) -> list[str]:
        """Get the column names for this measure.

        Returns
        -------
        list[str]
            The column names for this measure.
        """
        return ["Impact"]


class AssetStressTestMeasureSettings(MeasureSettings):
    """Settings for asset stress test measure.

    This class defines settings for an asset stress test measure,
    which provides asset stress test calculations.
    """

    type: Literal["AssetStressTest"] = Field("AssetStressTest", repr=False)
    shock_size: float = Field(
        1.0, description="The size of the shock to apply in standard deviations."
    )

    @property
    def columns(self) -> list[str]:
        """Get the column names for this measure.

        Returns
        -------
        list[str]
            The column names for this measure.
        """
        return ["Impact"]


class AssetHoldingsMeasureSettings(MeasureSettings):
    """Settings for asset holdings measure.

    This class defines settings for an asset holdings measure,
    which provides asset holdings calculations and comparisons.
    """

    type: Literal["AssetHoldings"] = Field("AssetHoldings", repr=False)

    @property
    def columns(self) -> list[str]:
        """Get the column names for this measure.

        Returns
        -------
        list[str]
            The column names for this measure.
        """
        return [
            *["Position", "Position Bench", "Position Diff"],
            *["Holdings", "Holdings Bench", "Holdings Diff"],
            *["Weight", "Weight Bench", "Weight Diff"],
            *["Rank", "Rank Bench", "Rank Diff"],
        ]


MeasureSettingsType: TypeAlias = Union[tuple(MeasureSettings.__subclasses__())]  # type: ignore
MeasureSettingsClassType = Type[MeasureSettingsType]

ALL_HOLDINGS_MEASURE_TYPES = [
    HoldingsMeasureSettings,
]
ALL_EXPOSURE_MEASURE_TYPES = [
    ExposureMeasureSettings,
]
ALL_XSR_MEASURE_TYPES = [
    XSigmaRhoMeasureSettings,
]
ALL_TSXSR_MEASURE_TYPES = [
    TimeSeriesXSigmaRhoMeasureSettings,
]
ALL_RISK_DECOMPOSITION_MEASURE_TYPES = [
    RiskDecompositionMeasureSettings,
]
ALL_TS_BETA_MEASURE_TYPES = [
    TimeSeriesBetaMeasureSettings,
]
ALL_VAR_MEASURE_TYPES = [
    HVaRMeasureSettings,
    AVaRMeasureSettings,
]
ALL_FACTOR_TS_MEASURE_TYPES = [
    TimeSeriesVolatilityMeasureSettings,
    CumsumMeasureSettings,
    DrawdownMeasureSettings,
]
ALL_RISK_MODEL_FIT_MEASURE_TYPES = [
    MovingAverageMeasureSettings,
]
ALL_RISK_MODEL_FIT_FACTOR_MEASURE_TYPES = [
    FactorMovingAverageMeasureSettings,
]
ALL_RISK_MODEL_PORTFOLIO_FIT_MEASURE_TYPES = [
    MovingAverageRSquaredMeasureSettings,
]
ALL_FACTOR_FORECAST_LOSS_MEASURE_TYPES = [
    FactorVolForecastMeasureSettings,
]
ALL_FORECAST_LOSS_MEASURE_TYPES = [
    VolForecastMeasureSettings,
]
ALL_FACTOR_COVARIANCE_MEASURE_TYPES = [
    FactorCovarianceMeasureSettings,
]
ALL_IDIOSYNCRATIC_VOLATILITY_MEASURE_TYPES = [
    IdiosyncraticVolatilityMeasureSettings,
]
ALL_IDIOSYNCRATIC_RETURN_MEASURE_TYPES = [
    IdiosyncraticReturnMeasureSettings,
]
ALL_FACTOR_IDIO_MEASURE_TYPES = [
    FactorIdioMeasureSettings,
]
ALL_STYLE_CORRELATION_MEASURE_TYPES = [
    PassThroughFactor2DMeasureSettings,
]
ALL_STYLE_INDUSTRY_EXPOSURE_MEASURE_TYPES = [
    PassThroughFactor2DMeasureSettings,
]
ALL_FORECAST_BACKTEST_MEASURE_TYPES = [
    ForecastBacktestMeasureSettings,
]
ALL_BRINSON_ATTRIBUTION_MEASURE_TYPES = [
    BrinsonAttributionMeasureSettings,
]
ALL_FACTOR_ATTRIBUTION_MEASURE_TYPES = [
    FactorAttributionMeasureSettings,
]
ALL_FACTOR_ATTRIBUTION_DRILLDOWN_MEASURE_TYPES = [
    FactorAttributionDrilldownMeasureSettings,
]
ALL_PORTFOLIO_STRESS_TEST_MEASURE_TYPES = [
    PortfolioStressTestMeasureSettings,
]
ALL_ASSET_STRESS_TEST_MEASURE_TYPES = [
    AssetStressTestMeasureSettings,
]
ALL_ASSET_HOLDINGS_MEASURE_TYPES = [
    AssetHoldingsMeasureSettings,
]


def _get_default_measures(
    measure_types: list[MeasureSettingsClassType],
) -> list[MeasureSettingsType]:
    return [m() for m in measure_types]


class ConcreteReportSettingsMenu(SettingsMenu, ABC):
    """Abstract base class for concrete report settings menus.

    This class defines the interface for concrete report settings menus
    that provide specific measure types for different report categories.
    """

    measure_types: ClassVar[list[MeasureSettingsClassType]] = []

    def describe(self) -> str:
        """Describe the available report settings.

        Returns
        -------
        str
            A description of the available report settings.
        """
        # TODO implement
        return "NA"


class HoldingsReportSettingsMenu(ConcreteReportSettingsMenu):
    """Settings menu for holdings reports.

    This class defines the available measure types for holdings reports.
    """

    measure_types: ClassVar[list[MeasureSettingsClassType]] = ALL_HOLDINGS_MEASURE_TYPES


class ExposureReportSettingsMenu(ConcreteReportSettingsMenu):
    """Settings menu for exposure reports.

    This class defines the available measure types for exposure reports.
    """

    measure_types: ClassVar[list[MeasureSettingsClassType]] = ALL_EXPOSURE_MEASURE_TYPES


class XSRReportSettingsMenu(ConcreteReportSettingsMenu):
    """Settings menu for X-Sigma-Rho reports.

    This class defines the available measure types for X-Sigma-Rho reports.
    """

    measure_types: ClassVar[list[MeasureSettingsClassType]] = ALL_XSR_MEASURE_TYPES


class TSXSRReportSettingsMenu(ConcreteReportSettingsMenu):
    """Settings menu for time series X-Sigma-Rho reports.

    This class defines the available measure types for time series X-Sigma-Rho reports.
    """

    measure_types: ClassVar[list[MeasureSettingsClassType]] = ALL_TSXSR_MEASURE_TYPES


class RiskDecompositionReportSettingsMenu(ConcreteReportSettingsMenu):
    """Settings menu for a risk decomposition reports.

    This class defines the available measure types for a risk decomposition reports.
    """

    measure_types: ClassVar[list[MeasureSettingsClassType]] = (
        ALL_RISK_DECOMPOSITION_MEASURE_TYPES
    )


class TSBetaReportSettingsMenu(ConcreteReportSettingsMenu):
    """Settings menu for time series beta reports.

    This class defines the available measure types for time series beta reports.
    """

    measure_types: ClassVar[list[MeasureSettingsClassType]] = ALL_TS_BETA_MEASURE_TYPES


class VaRReportSettingsMenu(ConcreteReportSettingsMenu):
    """Settings menu for VaR reports.

    This class defines the available measure types for VaR reports.
    """

    measure_types: ClassVar[list[MeasureSettingsClassType]] = ALL_VAR_MEASURE_TYPES


class FactorTSReportSettingsMenu(ConcreteReportSettingsMenu):
    """Settings menu for factor time series reports.

    This class defines the available measure types for factor time series reports.
    """

    measure_types: ClassVar[list[MeasureSettingsClassType]] = (
        ALL_FACTOR_TS_MEASURE_TYPES  # type: ignore
    )


class RiskModelFitReportSettingsMenu(ConcreteReportSettingsMenu):
    """Settings menu for risk model fit reports.

    This class defines the available measure types for risk model fit reports.
    """

    measure_types: ClassVar[list[MeasureSettingsClassType]] = (
        ALL_RISK_MODEL_FIT_MEASURE_TYPES
    )


class RiskModelFitFactorReportSettingsMenu(ConcreteReportSettingsMenu):
    """Settings menu for risk model fit factor reports.

    This class defines the available measure types for risk model fit factor reports.
    """

    measure_types: ClassVar[list[MeasureSettingsClassType]] = (
        ALL_RISK_MODEL_FIT_FACTOR_MEASURE_TYPES
    )


class RiskModelPortfolioFitReportSettingsMenu(ConcreteReportSettingsMenu):
    """Settings menu for risk model portfolio fit reports.

    This class defines the available measure types for risk model portfolio fit reports.
    """

    measure_types: ClassVar[list[MeasureSettingsClassType]] = (
        ALL_RISK_MODEL_FIT_MEASURE_TYPES
    )


class FactorForecastLossReportSettingsMenu(ConcreteReportSettingsMenu):
    """Settings menu for factor forecast loss reports.

    This class defines the available measure types for factor forecast loss reports.
    """

    measure_types: ClassVar[list[MeasureSettingsClassType]] = (
        ALL_FORECAST_LOSS_MEASURE_TYPES
    )


class AssetForecastLossReportSettingsMenu(ConcreteReportSettingsMenu):
    """Settings menu for asset forecast loss reports.

    This class defines the available measure types for asset forecast loss reports.
    """

    measure_types: ClassVar[list[MeasureSettingsClassType]] = (
        ALL_FORECAST_LOSS_MEASURE_TYPES
    )


class FactorCovarianceReportSettingsMenu(ConcreteReportSettingsMenu):
    """Settings menu for factor covariance reports.

    This class defines the available measure types for factor covariance reports.
    """

    measure_types: ClassVar[list[MeasureSettingsClassType]] = (
        ALL_FACTOR_COVARIANCE_MEASURE_TYPES
    )


class IdiosyncraticReturnReportSettingsMenu(ConcreteReportSettingsMenu):
    """Settings menu for idiosyncratic return reports.

    This class defines the available measure types for idiosyncratic return reports.
    """

    measure_types: ClassVar[list[MeasureSettingsClassType]] = (
        ALL_IDIOSYNCRATIC_RETURN_MEASURE_TYPES
    )


class IdiosyncraticVolatilityReportSettingsMenu(ConcreteReportSettingsMenu):
    """Settings menu for idiosyncratic volatility reports.

    This class defines the available measure types for idiosyncratic volatility reports.
    """

    measure_types: ClassVar[list[MeasureSettingsClassType]] = (
        ALL_IDIOSYNCRATIC_VOLATILITY_MEASURE_TYPES
    )


class FactorIdioReportSettingsMenu(ConcreteReportSettingsMenu):
    """Settings menu for factor idiosyncratic reports.

    This class defines the available measure types for factor idiosyncratic reports.
    """

    measure_types: ClassVar[list[MeasureSettingsClassType]] = (
        ALL_FACTOR_IDIO_MEASURE_TYPES
    )


class StyleCorrelationReportSettingsMenu(ConcreteReportSettingsMenu):
    """Settings menu for style correlation reports.

    This class defines the available measure types for style correlation reports.
    """

    measure_types: ClassVar[list[MeasureSettingsClassType]] = (
        ALL_STYLE_CORRELATION_MEASURE_TYPES
    )


class StyleIndustryExposureReportSettingsMenu(ConcreteReportSettingsMenu):
    """Settings menu for style industry exposure reports.

    This class defines the available measure types for style industry exposure reports.
    """

    measure_types: ClassVar[list[MeasureSettingsClassType]] = (
        ALL_STYLE_INDUSTRY_EXPOSURE_MEASURE_TYPES
    )


class ForecastBacktestReportSettingsMenu(ConcreteReportSettingsMenu):
    """Settings menu for forecast backtest reports.

    This class defines the available measure types for forecast backtest reports.
    """

    measure_types: ClassVar[list[MeasureSettingsClassType]] = (
        ALL_FORECAST_BACKTEST_MEASURE_TYPES
    )


class BrinsonAttributionReportSettingsMenu(ConcreteReportSettingsMenu):
    """Settings menu for Brinson attribution reports.

    This class defines the available measure types for Brinson attribution reports.
    """

    measure_types: ClassVar[list[MeasureSettingsClassType]] = (
        ALL_BRINSON_ATTRIBUTION_MEASURE_TYPES
    )


class FactorAttributionReportSettingsMenu(ConcreteReportSettingsMenu):
    """Settings menu for factor attribution reports.

    This class defines the available measure types for factor attribution reports.
    """

    measure_types: ClassVar[list[MeasureSettingsClassType]] = (
        ALL_FACTOR_ATTRIBUTION_MEASURE_TYPES
    )


class FactorAttributionDrilldownReportSettingsMenu(ConcreteReportSettingsMenu):
    """Settings menu for factor attribution drilldown reports.

    This class defines the available measure types for factor attribution drilldown
    reports.
    """

    measure_types: ClassVar[list[MeasureSettingsClassType]] = (
        ALL_FACTOR_ATTRIBUTION_DRILLDOWN_MEASURE_TYPES
    )


class PortfolioStressTestReportSettingsMenu(ConcreteReportSettingsMenu):
    """Settings menu for portfolio stress test reports.

    This class defines the available measure types for portfolio stress test reports.
    """

    measure_types: ClassVar[list[MeasureSettingsClassType]] = (
        ALL_PORTFOLIO_STRESS_TEST_MEASURE_TYPES
    )


class AssetStressTestReportSettingsMenu(ConcreteReportSettingsMenu):
    """Settings menu for asset stress test reports.

    This class defines the available measure types for asset stress test reports.
    """

    measure_types: ClassVar[list[MeasureSettingsClassType]] = (
        ALL_ASSET_STRESS_TEST_MEASURE_TYPES
    )


class AssetHoldingsReportSettingsMenu(ConcreteReportSettingsMenu):
    """Settings menu for asset holdings reports.

    This class defines the available measure types for asset holdings reports.
    """

    measure_types: ClassVar[list[MeasureSettingsClassType]] = (
        ALL_ASSET_HOLDINGS_MEASURE_TYPES
    )


class ReportSettingsMenu(SettingsMenu):
    """Defines available report settings to build a report."""

    hld: ConcreteReportSettingsMenu = HoldingsReportSettingsMenu()
    exp: ConcreteReportSettingsMenu = ExposureReportSettingsMenu()
    xsr: ConcreteReportSettingsMenu = XSRReportSettingsMenu()
    tsx: ConcreteReportSettingsMenu = TSXSRReportSettingsMenu()
    rdc: ConcreteReportSettingsMenu = RiskDecompositionReportSettingsMenu()
    bet: ConcreteReportSettingsMenu = TSBetaReportSettingsMenu()
    var: ConcreteReportSettingsMenu = VaRReportSettingsMenu()
    fts: ConcreteReportSettingsMenu = FactorTSReportSettingsMenu()
    rmf: ConcreteReportSettingsMenu = RiskModelFitReportSettingsMenu()
    rms: ConcreteReportSettingsMenu = RiskModelFitFactorReportSettingsMenu()
    rmp: ConcreteReportSettingsMenu = RiskModelPortfolioFitReportSettingsMenu()
    ffl: ConcreteReportSettingsMenu = FactorForecastLossReportSettingsMenu()
    afl: ConcreteReportSettingsMenu = AssetForecastLossReportSettingsMenu()
    fcr: ConcreteReportSettingsMenu = FactorCovarianceReportSettingsMenu()
    irr: ConcreteReportSettingsMenu = IdiosyncraticReturnReportSettingsMenu()
    ivr: ConcreteReportSettingsMenu = IdiosyncraticVolatilityReportSettingsMenu()
    fni: ConcreteReportSettingsMenu = FactorIdioReportSettingsMenu()
    scr: ConcreteReportSettingsMenu = StyleCorrelationReportSettingsMenu()
    six: ConcreteReportSettingsMenu = StyleIndustryExposureReportSettingsMenu()
    fbt: ConcreteReportSettingsMenu = ForecastBacktestReportSettingsMenu()
    bat: ConcreteReportSettingsMenu = BrinsonAttributionReportSettingsMenu()
    fat: ConcreteReportSettingsMenu = FactorAttributionReportSettingsMenu()
    fad: ConcreteReportSettingsMenu = FactorAttributionDrilldownReportSettingsMenu()
    pst: ConcreteReportSettingsMenu = PortfolioStressTestReportSettingsMenu()
    ast: ConcreteReportSettingsMenu = AssetStressTestReportSettingsMenu()
    ahd: ConcreteReportSettingsMenu = AssetHoldingsReportSettingsMenu()

    def describe(self) -> str:
        """Describe the report settings.

        Returns
        -------
        str
            A description of the report settings menu.
        """
        return "NA"


class ReportSettings(Settings[ReportSettingsMenu]):
    """Defines settings to build a report."""

    report: ConcreteReportSettingsType
    risk_model: Annotated[
        str | int | FactorRiskModelSettings,
        Field(
            description="The risk model to use for the report.",
        ),
        SettingsTypeMetaData[str | int | FactorRiskModelSettings](
            references=FactorRiskModelSettings
        ),
    ]

    def describe(self, menu: ReportSettingsMenu) -> str:
        """Describe the report settings.

        Parameters
        ----------
        menu : ReportSettingsMenu
            The menu to get context information from.

        Returns
        -------
        str
            A description of the report settings.
        """
        return "NA"

    @property
    def menu_type(self) -> type[ReportSettingsMenu]:  # noqa: D102
        return ReportSettingsMenu


class ConcreteReportSettings(BaseModel, ABC, frozen=True, extra="forbid"):
    """Abstract base class for concrete report settings.

    This class defines the interface for concrete report settings
    that provide specific configuration for different report types.
    """

    report_type: str
    lagged_holdings: ClassVar[bool] = False
    measures: list[MeasureSettingsType] = Field(
        description="The measures to include in the report.",
    )


class HoldingsReportSettings(ConcreteReportSettings):
    """Defines settings to build a holdings report."""

    report_type: Literal["Holdings report"] = Field("Holdings report", repr=False)
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(ALL_HOLDINGS_MEASURE_TYPES),
        description="The measures to include in the report.",
    )

    @property
    def menu_type(self) -> type[HoldingsReportSettingsMenu]:  # noqa: D102
        return HoldingsReportSettingsMenu


class ExposureReportSettings(ConcreteReportSettings):
    """Defines settings to build an exposure report."""

    report_type: Literal["Exposure report"] = Field("Exposure report", repr=False)
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(ALL_EXPOSURE_MEASURE_TYPES),
        description="The measures to include in the report.",
    )

    @property
    def menu_type(self) -> type[ExposureReportSettingsMenu]:  # noqa: D102
        return ExposureReportSettingsMenu


class XSRReportSettings(ConcreteReportSettings):
    """Defines settings to build an XSR report."""

    report_type: Literal["XSR report"] = Field("XSR report", repr=False)
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(ALL_XSR_MEASURE_TYPES),
        description="The measures to include in the report.",
    )
    halflife_factor_vol: int = Field(
        42, description="The half-life for the factor volatility."
    )
    halflife_factor_vra: int | None = Field(
        None,
        description=(
            "The half-life for the factor volatility regime adjustment. "
            "If None, no adjustment is applied."
        ),
    )
    halflife_factor_cor: int = Field(
        126, description="The half-life for the factor correlation."
    )
    halflife_idio_vol: int = Field(
        42, description="The half-life for the idiosyncratic volatility."
    )
    shrink_factor_cor_method: Literal["LIS", "QIS", "GIS"] | None = Field(
        None,
        description=(
            "The method to use for the shrinkage of the factor correlation. If None, "
            "don't shrink."
        ),
    )
    shrink_factor_cor_length: int | None = Field(
        1008,
        description=(
            "If a shrinkage method is provided, the length of the shrinkage for the "
            "factor correlation."
        ),
    )
    shrink_factor_cor_standardized: bool = Field(
        False,
        description=(
            "If True, the correlation matrix is combined with an estimate based on the "
            "standardized returns. The halflife_factor_vra is used for the weights."
        ),
    )
    nw_lags_factor_vol: int = Field(
        0, description="The Newey-West lags of the factor volatility."
    )
    nw_lags_factor_vol_halflife_override: int | None = Field(
        None,
        description=(
            "The Newey-West lags override for the factor volatility for lagged returns."
        ),
    )
    nw_lags_factor_cor: int = Field(
        0, description="The Newey-West lags of the factor correlation."
    )
    nw_lags_idio_vol: int = Field(
        0, description="The Newey-West lags of the idiosyncratic volatility."
    )
    nw_lags_idio_vol_halflife_override: int | None = Field(
        None,
        description=(
            "The Newey-West lags override for the idiosyncratic volatility for lagged returns."
        ),
    )
    shrink_idio_vol: float = Field(
        0.0, description="The shrinkage factor for the idiosyncratic volatility."
    )
    clip_idio_return_lb: float | None = Field(
        -0.3,
        description=(
            "The lower bound to clip the idiosyncratic return in order to compute "
            "the idiosyncratic volatility. If None, no clipping is applied."
        ),
    )
    clip_idio_return_ub: float | None = Field(
        None,
        description=(
            "The upper bound to clip the idiosyncratic return in order to compute "
            "the idiosyncratic volatility. If None, no clipping is applied."
        ),
    )
    void_idio_return_lb: float | None = Field(
        None,
        description=(
            "The lower bound to remove the idiosyncratic return in order to compute "
            "the idiosyncratic volatility. Removed returns are treated as if there "
            "was a holiday for this asset. If None, no removal is applied. Voiding "
            "returns is applied before clipping."
        ),
    )
    void_idio_return_ub: float | None = Field(
        0.5,
        description=(
            "The upper bound to remove the idiosyncratic return in order to compute "
            "the idiosyncratic volatility. Removed returns are treated as if there "
            "was a holiday for this asset. If None, no removal is applied. Voiding "
            "returns is applied before clipping."
        ),
    )
    structural_model_fallback_halflife: int = Field(
        21,
        description=(
            "The half-life to use for the structural model fallback. The structural "
            "model computes idio vol based on the factor exposures. The halflife "
            "determines the dacay of this estimate when historical returns become "
            "available."
        ),
    )

    @property
    def menu_type(self) -> type[XSRReportSettingsMenu]:  # noqa: D102
        return XSRReportSettingsMenu


class TSXSRReportSettings(ConcreteReportSettings):
    """Defines settings to build a time-series XSR report."""

    report_type: Literal["TSXSR report"] = Field("TSXSR report", repr=False)
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(ALL_TSXSR_MEASURE_TYPES),
        description="The measures to include in the report.",
    )
    halflife_factor_vol: int = Field(
        42, description="The half-life for the factor volatility."
    )
    halflife_factor_vra: int | None = Field(
        None,
        description=(
            "The half-life for the factor volatility regime adjustment. "
            "If None, no adjustment is applied."
        ),
    )
    halflife_factor_cor: int = Field(
        126, description="The half-life for the factor correlation."
    )
    halflife_idio_vol: int = Field(
        42, description="The half-life for the idiosyncratic volatility."
    )
    shrink_factor_cor_method: Literal["LIS", "QIS", "GIS"] | None = Field(
        None,
        description=(
            "The method to use for the shrinkage of the factor correlation. If None, "
            "don't shrink."
        ),
    )
    shrink_factor_cor_length: int | None = Field(
        1008,
        description=(
            "If a shrinkage method is provided, the length of the shrinkage for the "
            "factor correlation."
        ),
    )
    shrink_factor_cor_standardized: bool = Field(
        False,
        description=(
            "If True, the correlation matrix is combined with an estimate based on the "
            "standardized returns. The halflife_factor_vra is used for the weights."
        ),
    )
    nw_lags_factor_vol: int = Field(
        0, description="The Newey-West lags of the factor volatility."
    )
    nw_lags_factor_vol_halflife_override: int | None = Field(
        None,
        description=(
            "The Newey-West lags override for the factor volatility for lagged returns."
        ),
    )
    nw_lags_factor_cor: int = Field(
        0, description="The Newey-West lags of the factor correlation."
    )
    nw_lags_idio_vol: int = Field(
        0, description="The Newey-West lags of the idiosyncratic volatility."
    )
    nw_lags_idio_vol_halflife_override: int | None = Field(
        None,
        description=(
            "The Newey-West lags override for the idiosyncratic volatility for lagged returns."
        ),
    )
    shrink_idio_vol: float = Field(
        0.0, description="The shrinkage factor for the idiosyncratic volatility."
    )
    clip_idio_return_lb: float | None = Field(
        -0.3,
        description=(
            "The lower bound to clip the idiosyncratic return in order to compute "
            "the idiosyncratic volatility. If None, no clipping is applied."
        ),
    )
    clip_idio_return_ub: float | None = Field(
        None,
        description=(
            "The upper bound to clip the idiosyncratic return in order to compute "
            "the idiosyncratic volatility. If None, no clipping is applied."
        ),
    )
    void_idio_return_lb: float | None = Field(
        None,
        description=(
            "The lower bound to remove the idiosyncratic return in order to compute "
            "the idiosyncratic volatility. Removed returns are treated as if there "
            "was a holiday for this asset. If None, no removal is applied. Voiding "
            "returns is applied before clipping."
        ),
    )
    void_idio_return_ub: float | None = Field(
        0.5,
        description=(
            "The upper bound to remove the idiosyncratic return in order to compute "
            "the idiosyncratic volatility. Removed returns are treated as if there "
            "was a holiday for this asset. If None, no removal is applied. Voiding "
            "returns is applied before clipping."
        ),
    )
    structural_model_fallback_halflife: int = Field(
        21,
        description=(
            "The half-life to use for the structural model fallback. The structural "
            "model computes idio vol based on the factor exposures. The halflife "
            "determines the dacay of this estimate when historical returns become "
            "available."
        ),
    )


class RiskDecompositionReportSettings(ConcreteReportSettings):
    """Defines settings to build a risk decomposition report."""

    report_type: Literal["Risk decomposition report"] = Field(
        "Risk decomposition report", repr=False
    )
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(
            ALL_RISK_DECOMPOSITION_MEASURE_TYPES
        ),
        description="The measures to include in the report.",
    )
    halflife_factor_vol: int = Field(
        42, description="The half-life for the factor volatility."
    )
    halflife_factor_vra: int | None = Field(
        None,
        description=(
            "The half-life for the factor volatility regime adjustment. "
            "If None, no adjustment is applied."
        ),
    )
    halflife_factor_cor: int = Field(
        126, description="The half-life for the factor correlation."
    )
    halflife_idio_vol: int = Field(
        42, description="The half-life for the idiosyncratic volatility."
    )
    shrink_factor_cor_method: Literal["LIS", "QIS", "GIS"] | None = Field(
        None,
        description=(
            "The method to use for the shrinkage of the factor correlation. If None, "
            "don't shrink."
        ),
    )
    shrink_factor_cor_length: int | None = Field(
        1008,
        description=(
            "If a shrinkage method is provided, the length of the shrinkage for the "
            "factor correlation."
        ),
    )
    shrink_factor_cor_standardized: bool = Field(
        False,
        description=(
            "If True, the correlation matrix is combined with an estimate based on the "
            "standardized returns. The halflife_factor_vra is used for the weights."
        ),
    )
    nw_lags_factor_vol: int = Field(
        0, description="The Newey-West lags of the factor volatility."
    )
    nw_lags_factor_vol_halflife_override: int | None = Field(
        None,
        description=(
            "The Newey-West lags override for the factor volatility for lagged returns."
        ),
    )
    nw_lags_factor_cor: int = Field(
        0, description="The Newey-West lags of the factor correlation."
    )
    nw_lags_idio_vol: int = Field(
        0, description="The Newey-West lags of the idiosyncratic volatility."
    )
    nw_lags_idio_vol_halflife_override: int | None = Field(
        None,
        description=(
            "The Newey-West lags override for the idiosyncratic volatility for lagged returns."
        ),
    )
    shrink_idio_vol: float = Field(
        0.0, description="The shrinkage factor for the idiosyncratic volatility."
    )
    clip_idio_return_lb: float | None = Field(
        -0.3,
        description=(
            "The lower bound to clip the idiosyncratic return in order to compute "
            "the idiosyncratic volatility. If None, no clipping is applied."
        ),
    )
    clip_idio_return_ub: float | None = Field(
        None,
        description=(
            "The upper bound to clip the idiosyncratic return in order to compute "
            "the idiosyncratic volatility. If None, no clipping is applied."
        ),
    )
    void_idio_return_lb: float | None = Field(
        None,
        description=(
            "The lower bound to remove the idiosyncratic return in order to compute "
            "the idiosyncratic volatility. Removed returns are treated as if there "
            "was a holiday for this asset. If None, no removal is applied. Voiding "
            "returns is applied before clipping."
        ),
    )
    void_idio_return_ub: float | None = Field(
        0.5,
        description=(
            "The upper bound to remove the idiosyncratic return in order to compute "
            "the idiosyncratic volatility. Removed returns are treated as if there "
            "was a holiday for this asset. If None, no removal is applied. Voiding "
            "returns is applied before clipping."
        ),
    )
    structural_model_fallback_halflife: int = Field(
        21,
        description=(
            "The half-life to use for the structural model fallback. The structural "
            "model computes idio vol based on the factor exposures. The halflife "
            "determines the dacay of this estimate when historical returns become "
            "available."
        ),
    )


class TSBetaReportSettings(ConcreteReportSettings):
    """Defines settings to build a time-series Beta report."""

    report_type: Literal["TS Beta report"] = Field("TS Beta report", repr=False)
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(ALL_TS_BETA_MEASURE_TYPES),
        description="The measures to include in the report.",
    )
    halflife_factor_vol: int = Field(
        42, description="The half-life for the factor volatility."
    )
    halflife_factor_vra: int | None = Field(
        None,
        description=(
            "The half-life for the factor volatility regime adjustment. "
            "If None, no adjustment is applied."
        ),
    )
    halflife_factor_cor: int = Field(
        126, description="The half-life for the factor correlation."
    )
    halflife_idio_vol: int = Field(
        42, description="The half-life for the idiosyncratic volatility."
    )
    shrink_factor_cor_method: Literal["LIS", "QIS", "GIS"] | None = Field(
        None,
        description=(
            "The method to use for the shrinkage of the factor correlation. If None, "
            "don't shrink."
        ),
    )
    shrink_factor_cor_length: int | None = Field(
        1008,
        description=(
            "If a shrinkage method is provided, the length of the shrinkage for the "
            "factor correlation."
        ),
    )
    shrink_factor_cor_standardized: bool = Field(
        False,
        description=(
            "If True, the correlation matrix is combined with an estimate based on the "
            "standardized returns. The halflife_factor_vra is used for the weights."
        ),
    )
    nw_lags_factor_vol: int = Field(
        0, description="The Newey-West lags of the factor volatility."
    )
    nw_lags_factor_vol_halflife_override: int | None = Field(
        None,
        description=(
            "The Newey-West lags override for the factor volatility for lagged returns."
        ),
    )
    nw_lags_factor_cor: int = Field(
        0, description="The Newey-West lags of the factor correlation."
    )
    nw_lags_idio_vol: int = Field(
        0, description="The Newey-West lags of the idiosyncratic volatility."
    )
    nw_lags_idio_vol_halflife_override: int | None = Field(
        None,
        description=(
            "The Newey-West lags override for the idiosyncratic volatility for lagged returns."
        ),
    )
    shrink_idio_vol: float = Field(
        0.0, description="The shrinkage factor for the idiosyncratic volatility."
    )
    clip_idio_return_lb: float | None = Field(
        -0.3,
        description=(
            "The lower bound to clip the idiosyncratic return in order to compute "
            "the idiosyncratic volatility. If None, no clipping is applied."
        ),
    )
    clip_idio_return_ub: float | None = Field(
        None,
        description=(
            "The upper bound to clip the idiosyncratic return in order to compute "
            "the idiosyncratic volatility. If None, no clipping is applied."
        ),
    )
    void_idio_return_lb: float | None = Field(
        None,
        description=(
            "The lower bound to remove the idiosyncratic return in order to compute "
            "the idiosyncratic volatility. Removed returns are treated as if there "
            "was a holiday for this asset. If None, no removal is applied. Voiding "
            "returns is applied before clipping."
        ),
    )
    void_idio_return_ub: float | None = Field(
        0.5,
        description=(
            "The upper bound to remove the idiosyncratic return in order to compute "
            "the idiosyncratic volatility. Removed returns are treated as if there "
            "was a holiday for this asset. If None, no removal is applied. Voiding "
            "returns is applied before clipping."
        ),
    )
    structural_model_fallback_halflife: int = Field(
        21,
        description=(
            "The half-life to use for the structural model fallback. The structural "
            "model computes idio vol based on the factor exposures. The halflife "
            "determines the dacay of this estimate when historical returns become "
            "available."
        ),
    )

    @property
    def menu_type(self) -> type[TSXSRReportSettingsMenu]:  # noqa: D102
        return TSXSRReportSettingsMenu


class VaRReportSettings(ConcreteReportSettings):
    """Defines settings to build a VaR report."""

    report_type: Literal["VaR report"] = Field("VaR report", repr=False)
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(ALL_VAR_MEASURE_TYPES),
        description="The measures to include in the report.",
    )
    lookback: int = Field(1000, description="The lookback period for the estimation.")

    @property
    def menu_type(self) -> type[VaRReportSettingsMenu]:  # noqa: D102
        return VaRReportSettingsMenu


class FactorTSReportSettings(ConcreteReportSettings):
    """Defines settings to build a factor time-series report."""

    report_type: Literal["Factor TS report"] = Field("Factor TS report", repr=False)
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(ALL_FACTOR_TS_MEASURE_TYPES),  # type: ignore
        description="The measures to include in the report.",
    )

    @property
    def menu_type(self) -> type[FactorTSReportSettingsMenu]:  # noqa: D102
        return FactorTSReportSettingsMenu


class RiskModelFitReportSettings(ConcreteReportSettings):
    """Defines settings to build a risk model fit report."""

    report_type: Literal["Risk Model Fit report"] = Field(
        "Risk Model Fit report", repr=False
    )
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(ALL_RISK_MODEL_FIT_MEASURE_TYPES),
        description="The measures to include in the report.",
    )

    @property
    def menu_type(self) -> type[RiskModelFitReportSettingsMenu]:  # noqa: D102
        return RiskModelFitReportSettingsMenu


class RiskModelFitFactorReportSettings(ConcreteReportSettings):
    """Defines settings to build a risk model fit report at the factor level."""

    report_type: Literal["Risk Model Fit Factor report"] = Field(
        "Risk Model Fit Factor report", repr=False
    )
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(
            ALL_RISK_MODEL_FIT_FACTOR_MEASURE_TYPES
        ),
        description="The measures to include in the report.",
    )

    @property
    def menu_type(self) -> type[RiskModelFitFactorReportSettingsMenu]:  # noqa: D102
        return RiskModelFitFactorReportSettingsMenu


class RiskModelPortfolioFitReportSettings(ConcreteReportSettings):
    """Defines settings to build a risk model fit report at the portfolio level."""

    report_type: Literal["Risk Model Portfolio Fit report"] = Field(
        "Risk Model Portfolio Fit report", repr=False
    )
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(
            ALL_RISK_MODEL_PORTFOLIO_FIT_MEASURE_TYPES
        ),
        description="The measures to include in the report.",
    )

    @property
    def menu_type(self) -> type[RiskModelPortfolioFitReportSettingsMenu]:  # noqa: D102
        return RiskModelPortfolioFitReportSettingsMenu


class FactorForecastLossReportSettings(ConcreteReportSettings):
    """Defines settings to build a forecast loss report."""

    report_type: Literal["Factor Forecast Loss report"] = Field(
        "Factor Forecast Loss report", repr=False
    )
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(
            ALL_FACTOR_FORECAST_LOSS_MEASURE_TYPES
        ),
        description="The measures to include in the report.",
    )
    horizons: list[int] = Field([1, 5, 21], description="The forecast horizons to use.")
    halflife_factor_vol: int = Field(
        42, description="The half-life for the factor volatility."
    )
    halflife_factor_vra: int | None = Field(
        None,
        description=(
            "The half-life for the factor volatility regime adjustment. "
            "If None, no adjustment is applied."
        ),
    )
    nw_lags_factor_vol: int = Field(
        0, description="The Newey-West lags of the factor volatility."
    )
    nw_lags_factor_vol_halflife_override: int | None = Field(
        None,
        description=(
            "The Newey-West lags override for the factor volatility for lagged returns."
        ),
    )

    @property
    def menu_type(self) -> type[FactorForecastLossReportSettingsMenu]:  # noqa: D102
        return FactorForecastLossReportSettingsMenu


class AssetForecastLossReportSettings(ConcreteReportSettings):
    """Defines settings to build a forecast loss report."""

    report_type: Literal["Forecast Loss report"] = Field(
        "Forecast Loss report", repr=False
    )
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(ALL_FORECAST_LOSS_MEASURE_TYPES),
        description="The measures to include in the report.",
    )
    horizons: list[int] = Field([1, 5, 21], description="The forecast horizons to use.")
    halflife_factor_vol: int = Field(
        42, description="The half-life for the factor volatility."
    )
    halflife_factor_vra: int | None = Field(
        None,
        description=(
            "The half-life for the factor volatility regime adjustment. "
            "If None, no adjustment is applied."
        ),
    )
    halflife_factor_cor: int = Field(
        126, description="The half-life for the factor correlation."
    )
    halflife_idio_vol: int = Field(
        42, description="The half-life for the idiosyncratic volatility."
    )
    shrink_factor_cor_method: Literal["LIS", "QIS", "GIS"] | None = Field(
        None,
        description=(
            "The method to use for the shrinkage of the factor correlation. If None, "
            "don't shrink."
        ),
    )
    shrink_factor_cor_length: int | None = Field(
        1008,
        description=(
            "If a shrinkage method is provided, the length of the shrinkage for the "
            "factor correlation."
        ),
    )
    shrink_factor_cor_standardized: bool = Field(
        False,
        description=(
            "If True, the correlation matrix is combined with an estimate based on the "
            "standardized returns. The halflife_factor_vra is used for the weights."
        ),
    )
    nw_lags_factor_vol: int = Field(
        0, description="The Newey-West lags of the factor volatility."
    )
    nw_lags_factor_vol_halflife_override: int | None = Field(
        None,
        description=(
            "The Newey-West lags override for the factor volatility for lagged returns."
        ),
    )
    nw_lags_factor_cor: int = Field(
        0, description="The Newey-West lags of the factor correlation."
    )
    nw_lags_idio_vol: int = Field(
        0, description="The Newey-West lags of the idiosyncratic volatility."
    )
    nw_lags_idio_vol_halflife_override: int | None = Field(
        None,
        description=(
            "The Newey-West lags override for the idiosyncratic volatility for lagged returns."
        ),
    )
    shrink_idio_vol: float = Field(
        0.0, description="The shrinkage factor for the idiosyncratic volatility."
    )
    clip_idio_return_lb: float | None = Field(
        -0.3,
        description=(
            "The lower bound to clip the idiosyncratic return in order to compute "
            "the idiosyncratic volatility. If None, no clipping is applied."
        ),
    )
    clip_idio_return_ub: float | None = Field(
        None,
        description=(
            "The upper bound to clip the idiosyncratic return in order to compute "
            "the idiosyncratic volatility. If None, no clipping is applied."
        ),
    )
    void_idio_return_lb: float | None = Field(
        None,
        description=(
            "The lower bound to remove the idiosyncratic return in order to compute "
            "the idiosyncratic volatility. Removed returns are treated as if there "
            "was a holiday for this asset. If None, no removal is applied. Voiding "
            "returns is applied before clipping."
        ),
    )
    void_idio_return_ub: float | None = Field(
        0.5,
        description=(
            "The upper bound to remove the idiosyncratic return in order to compute "
            "the idiosyncratic volatility. Removed returns are treated as if there "
            "was a holiday for this asset. If None, no removal is applied. Voiding "
            "returns is applied before clipping."
        ),
    )
    structural_model_fallback_halflife: int = Field(
        21,
        description=(
            "The half-life to use for the structural model fallback. The structural "
            "model computes idio vol based on the factor exposures. The halflife "
            "determines the dacay of this estimate when historical returns become "
            "available."
        ),
    )

    @property
    def menu_type(self) -> type[AssetForecastLossReportSettingsMenu]:  # noqa: D102
        return AssetForecastLossReportSettingsMenu


class FactorCovarianceReportSettings(ConcreteReportSettings):
    """Defines settings to build a factor covariance report."""

    report_type: Literal["Factor Covariance report"] = Field(
        "Factor Covariance report", repr=False
    )
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(
            ALL_FACTOR_COVARIANCE_MEASURE_TYPES
        ),
        description="The measures to include in the report.",
    )
    halflife_factor_vol: int = Field(
        42, description="The half-life for the factor volatility."
    )
    halflife_factor_vra: int | None = Field(
        None,
        description=(
            "The half-life for the factor volatility regime adjustment. "
            "If None, no adjustment is applied."
        ),
    )
    halflife_factor_cor: int = Field(
        126, description="The half-life for the factor correlation."
    )
    shrink_factor_cor_method: Literal["LIS", "QIS", "GIS"] | None = Field(
        None,
        description=(
            "The method to use for the shrinkage of the factor correlation. If None, "
            "don't shrink."
        ),
    )
    shrink_factor_cor_length: int | None = Field(
        1008,
        description=(
            "If a shrinkage method is provided, the length of the shrinkage for the "
            "factor correlation."
        ),
    )
    shrink_factor_cor_standardized: bool = Field(
        False,
        description=(
            "If True, the correlation matrix is combined with an estimate based on the "
            "standardized returns. The halflife_factor_vra is used for the weights."
        ),
    )
    nw_lags_factor_vol: int = Field(
        0, description="The Newey-West lags of the factor volatility."
    )
    nw_lags_factor_vol_halflife_override: int | None = Field(
        None,
        description=(
            "The Newey-West lags override for the factor volatility for lagged returns."
        ),
    )
    nw_lags_factor_cor: int = Field(
        0, description="The Newey-West lags of the factor correlation."
    )

    @property
    def menu_type(self) -> type[FactorCovarianceReportSettingsMenu]:  # noqa: D102
        return FactorCovarianceReportSettingsMenu


class IdiosyncraticReturnReportSettings(ConcreteReportSettings):
    """Defines settings to build an idiosyncratic return report."""

    report_type: Literal["Idiosyncratic Return report"] = Field(
        "Idiosyncratic Return report", repr=False
    )
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(
            ALL_IDIOSYNCRATIC_RETURN_MEASURE_TYPES
        ),
        description="The measures to include in the report.",
    )

    @property
    def menu_type(self) -> type[IdiosyncraticReturnReportSettingsMenu]:  # noqa: D102
        return IdiosyncraticReturnReportSettingsMenu


class IdiosyncraticVolatilityReportSettings(ConcreteReportSettings):
    """Defines settings to build an idiosyncratic volatility report."""

    report_type: Literal["Idiosyncratic Volatility report"] = Field(
        "Idiosyncratic Volatility report", repr=False
    )
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(
            ALL_IDIOSYNCRATIC_VOLATILITY_MEASURE_TYPES
        ),
        description="The measures to include in the report.",
    )
    halflife_idio_vol: int = Field(
        42, description="The half-life for the idiosyncratic volatility."
    )
    nw_lags_idio_vol: int = Field(
        0, description="The Newey-West lags of the idiosyncratic volatility."
    )
    nw_lags_idio_vol_halflife_override: int | None = Field(
        None,
        description=(
            "The Newey-West lags override for the idiosyncratic volatility for lagged returns."
        ),
    )
    shrink_idio_vol: float = Field(
        0.0, description="The shrinkage factor for the idiosyncratic volatility."
    )
    clip_idio_return_lb: float | None = Field(
        -0.3,
        description=(
            "The lower bound to clip the idiosyncratic return in order to compute "
            "the idiosyncratic volatility. If None, no clipping is applied."
        ),
    )
    clip_idio_return_ub: float | None = Field(
        None,
        description=(
            "The upper bound to clip the idiosyncratic return in order to compute "
            "the idiosyncratic volatility. If None, no clipping is applied."
        ),
    )
    void_idio_return_lb: float | None = Field(
        None,
        description=(
            "The lower bound to remove the idiosyncratic return in order to compute "
            "the idiosyncratic volatility. Removed returns are treated as if there "
            "was a holiday for this asset. If None, no removal is applied. Voiding "
            "returns is applied before clipping."
        ),
    )
    void_idio_return_ub: float | None = Field(
        0.5,
        description=(
            "The upper bound to remove the idiosyncratic return in order to compute "
            "the idiosyncratic volatility. Removed returns are treated as if there "
            "was a holiday for this asset. If None, no removal is applied. Voiding "
            "returns is applied before clipping."
        ),
    )
    structural_model_fallback_halflife: int = Field(
        21,
        description=(
            "The half-life to use for the structural model fallback. The structural "
            "model computes idio vol based on the factor exposures. The halflife "
            "determines the dacay of this estimate when historical returns become "
            "available."
        ),
    )

    @property
    def menu_type(  # noqa: D102
        self,
    ) -> type[IdiosyncraticVolatilityReportSettingsMenu]:
        return IdiosyncraticVolatilityReportSettingsMenu


class FactorIdioReportSettings(ConcreteReportSettings):
    """Defines settings to build a factor vcov and idio report."""

    report_type: Literal["Factor Idio report"] = Field("Factor Idio report", repr=False)
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(ALL_FACTOR_IDIO_MEASURE_TYPES),
        description="The measures to include in the report.",
    )
    halflife_factor_vol: int = Field(
        42, description="The half-life for the factor volatility."
    )
    halflife_factor_vra: int | None = Field(
        None,
        description=(
            "The half-life for the factor volatility regime adjustment. "
            "If None, no adjustment is applied."
        ),
    )
    halflife_factor_cor: int = Field(
        126, description="The half-life for the factor correlation."
    )
    halflife_idio_vol: int = Field(
        42, description="The half-life for the idiosyncratic volatility."
    )
    shrink_factor_cor_method: Literal["LIS", "QIS", "GIS"] | None = Field(
        None,
        description=(
            "The method to use for the shrinkage of the factor correlation. If None, "
            "don't shrink."
        ),
    )
    shrink_factor_cor_length: int | None = Field(
        1008,
        description=(
            "If a shrinkage method is provided, the length of the shrinkage for the "
            "factor correlation."
        ),
    )
    shrink_factor_cor_standardized: bool = Field(
        False,
        description=(
            "If True, the correlation matrix is combined with an estimate based on the "
            "standardized returns. The halflife_factor_vra is used for the weights."
        ),
    )
    nw_lags_factor_vol: int = Field(
        0, description="The Newey-West lags of the factor volatility."
    )
    nw_lags_factor_vol_halflife_override: int | None = Field(
        None,
        description=(
            "The Newey-West lags override for the factor volatility for lagged returns."
        ),
    )
    nw_lags_factor_cor: int = Field(
        0, description="The Newey-West lags of the factor correlation."
    )
    nw_lags_idio_vol: int = Field(
        0, description="The Newey-West lags of the idiosyncratic volatility."
    )
    nw_lags_idio_vol_halflife_override: int | None = Field(
        None,
        description=(
            "The Newey-West lags override for the idiosyncratic volatility for lagged returns."
        ),
    )
    shrink_idio_vol: float = Field(
        0.0, description="The shrinkage factor for the idiosyncratic volatility."
    )
    clip_idio_return_lb: float | None = Field(
        -0.3,
        description=(
            "The lower bound to clip the idiosyncratic return in order to compute "
            "the idiosyncratic volatility. If None, no clipping is applied."
        ),
    )
    clip_idio_return_ub: float | None = Field(
        None,
        description=(
            "The upper bound to clip the idiosyncratic return in order to compute "
            "the idiosyncratic volatility. If None, no clipping is applied."
        ),
    )
    void_idio_return_lb: float | None = Field(
        None,
        description=(
            "The lower bound to remove the idiosyncratic return in order to compute "
            "the idiosyncratic volatility. Removed returns are treated as if there "
            "was a holiday for this asset. If None, no removal is applied. Voiding "
            "returns is applied before clipping."
        ),
    )
    void_idio_return_ub: float | None = Field(
        0.5,
        description=(
            "The upper bound to remove the idiosyncratic return in order to compute "
            "the idiosyncratic volatility. Removed returns are treated as if there "
            "was a holiday for this asset. If None, no removal is applied. Voiding "
            "returns is applied before clipping."
        ),
    )
    structural_model_fallback_halflife: int = Field(
        21,
        description=(
            "The half-life to use for the structural model fallback. The structural "
            "model computes idio vol based on the factor exposures. The halflife "
            "determines the dacay of this estimate when historical returns become "
            "available."
        ),
    )

    @property
    def menu_type(self) -> type[FactorIdioReportSettingsMenu]:  # noqa: D102
        return FactorIdioReportSettingsMenu


class StyleCorrelationReportSettings(ConcreteReportSettings):
    """Defines settings to build a style correlation report."""

    report_type: Literal["Style Correlation report"] = Field(
        "Style Correlation report", repr=False
    )
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(
            ALL_STYLE_CORRELATION_MEASURE_TYPES
        ),
        description="The measures to include in the report.",
    )

    @property
    def menu_type(self) -> type[StyleCorrelationReportSettingsMenu]:  # noqa: D102
        return StyleCorrelationReportSettingsMenu


class StyleIndustryExposureReportSettings(ConcreteReportSettings):
    """Defines settings to build a style industry exposure report."""

    report_type: Literal["Style Industry Exposure report"] = Field(
        "Style Industry Exposure report", repr=False
    )
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(
            ALL_STYLE_INDUSTRY_EXPOSURE_MEASURE_TYPES
        ),
        description="The measures to include in the report.",
    )

    @property
    def menu_type(self) -> type[StyleIndustryExposureReportSettingsMenu]:  # noqa: D102
        return StyleIndustryExposureReportSettingsMenu


class ForecastBacktestReportSettings(ConcreteReportSettings):
    """Defines settings to build a forecast backtest report."""

    report_type: Literal["Forecast Backtest report"] = Field(
        "Forecast Backtest report", repr=False
    )
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(
            ALL_FORECAST_BACKTEST_MEASURE_TYPES
        ),
        description="The measures to include in the report.",
    )
    horizons: list[int] = Field([1, 5, 21], description="The forecast horizons to use.")
    halflife_factor_vol: int = Field(
        42, description="The half-life for the factor volatility."
    )
    halflife_factor_vra: int | None = Field(
        None,
        description=(
            "The half-life for the factor volatility regime adjustment. "
            "If None, no adjustment is applied."
        ),
    )
    halflife_factor_cor: int = Field(
        126, description="The half-life for the factor correlation."
    )
    halflife_idio_vol: int = Field(
        42, description="The half-life for the idiosyncratic volatility."
    )
    shrink_factor_cor_method: Literal["LIS", "QIS", "GIS"] | None = Field(
        None,
        description=(
            "The method to use for the shrinkage of the factor correlation. If None, "
            "don't shrink."
        ),
    )
    shrink_factor_cor_length: int | None = Field(
        1008,
        description=(
            "If a shrinkage method is provided, the length of the shrinkage for the "
            "factor correlation."
        ),
    )
    shrink_factor_cor_standardized: bool = Field(
        False,
        description=(
            "If True, the correlation matrix is combined with an estimate based on the "
            "standardized returns. The halflife_factor_vra is used for the weights."
        ),
    )
    nw_lags_factor_vol: int = Field(
        0, description="The Newey-West lags of the factor volatility."
    )
    nw_lags_factor_vol_halflife_override: int | None = Field(
        None,
        description=(
            "The Newey-West lags override for the factor volatility for lagged returns."
        ),
    )
    nw_lags_factor_cor: int = Field(
        0, description="The Newey-West lags of the factor correlation."
    )
    nw_lags_idio_vol: int = Field(
        0, description="The Newey-West lags of the idiosyncratic volatility."
    )
    nw_lags_idio_vol_halflife_override: int | None = Field(
        None,
        description=(
            "The Newey-West lags override for the idiosyncratic volatility for lagged returns."
        ),
    )
    shrink_idio_vol: float = Field(
        0.0, description="The shrinkage factor for the idiosyncratic volatility."
    )
    clip_idio_return_lb: float | None = Field(
        -0.3,
        description=(
            "The lower bound to clip the idiosyncratic return in order to compute "
            "the idiosyncratic volatility. If None, no clipping is applied."
        ),
    )
    clip_idio_return_ub: float | None = Field(
        None,
        description=(
            "The upper bound to clip the idiosyncratic return in order to compute "
            "the idiosyncratic volatility. If None, no clipping is applied."
        ),
    )
    void_idio_return_lb: float | None = Field(
        None,
        description=(
            "The lower bound to remove the idiosyncratic return in order to compute "
            "the idiosyncratic volatility. Removed returns are treated as if there "
            "was a holiday for this asset. If None, no removal is applied. Voiding "
            "returns is applied before clipping."
        ),
    )
    void_idio_return_ub: float | None = Field(
        0.5,
        description=(
            "The upper bound to remove the idiosyncratic return in order to compute "
            "the idiosyncratic volatility. Removed returns are treated as if there "
            "was a holiday for this asset. If None, no removal is applied. Voiding "
            "returns is applied before clipping."
        ),
    )
    structural_model_fallback_halflife: int = Field(
        21,
        description=(
            "The half-life to use for the structural model fallback. The structural "
            "model computes idio vol based on the factor exposures. The halflife "
            "determines the dacay of this estimate when historical returns become "
            "available."
        ),
    )

    @property
    def menu_type(self) -> type[ForecastBacktestReportSettingsMenu]:  # noqa: D102
        return ForecastBacktestReportSettingsMenu


class BrinsonAttributionReportSettings(ConcreteReportSettings):
    """Defines settings to build a Brinson attribution report."""

    report_type: Literal["Brinson Attribution report"] = Field(
        "Brinson Attribution report", repr=False
    )
    lagged_holdings: ClassVar[bool] = True
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(
            ALL_BRINSON_ATTRIBUTION_MEASURE_TYPES
        ),
        description="The measures to include in the report.",
    )

    @property
    def menu_type(self) -> type[BrinsonAttributionReportSettingsMenu]:  # noqa: D102
        return BrinsonAttributionReportSettingsMenu


class FactorAttributionReportSettings(ConcreteReportSettings):
    """Defines settings to build a factor attribution report."""

    report_type: Literal["Factor Attribution report"] = Field(
        "Factor Attribution report", repr=False
    )
    lagged_holdings: ClassVar[bool] = True
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(
            ALL_FACTOR_ATTRIBUTION_MEASURE_TYPES
        ),
        description="The measures to include in the report.",
    )

    @property
    def menu_type(self) -> type[FactorAttributionReportSettingsMenu]:  # noqa: D102
        return FactorAttributionReportSettingsMenu


class FactorAttributionDrilldownReportSettings(ConcreteReportSettings):
    """Defines settings to build a factor attribution drilldown report."""

    report_type: Literal["Factor Attribution Drilldown report"] = Field(
        "Factor Attribution Drilldown report", repr=False
    )
    lagged_holdings: ClassVar[bool] = True
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(
            ALL_FACTOR_ATTRIBUTION_DRILLDOWN_MEASURE_TYPES
        ),
        description="The measures to include in the report.",
    )

    @property
    def menu_type(  # noqa: D102
        self,
    ) -> type[FactorAttributionDrilldownReportSettingsMenu]:
        return FactorAttributionDrilldownReportSettingsMenu


class PortfolioStressTestReportSettings(ConcreteReportSettings):
    """Defines settings to build a portfolio stress test report."""

    report_type: Literal["Portfolio Stress Test report"] = Field(
        "Portfolio Stress Test report", repr=False
    )
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(
            ALL_PORTFOLIO_STRESS_TEST_MEASURE_TYPES
        ),
        description="The measures to include in the report.",
    )
    halflife_factor_vol: int = Field(
        42, description="The half-life for the factor volatility."
    )
    halflife_factor_vra: int | None = Field(
        None,
        description=(
            "The half-life for the factor volatility regime adjustment. "
            "If None, no adjustment is applied."
        ),
    )
    halflife_factor_cor: int = Field(
        126, description="The half-life for the factor correlation."
    )
    halflife_idio_vol: int = Field(
        42, description="The half-life for the idiosyncratic volatility."
    )
    shrink_factor_cor_method: Literal["LIS", "QIS", "GIS"] | None = Field(
        None,
        description=(
            "The method to use for the shrinkage of the factor correlation. If None, "
            "don't shrink."
        ),
    )
    shrink_factor_cor_length: int | None = Field(
        1008,
        description=(
            "If a shrinkage method is provided, the length of the shrinkage for the "
            "factor correlation."
        ),
    )
    shrink_factor_cor_standardized: bool = Field(
        False,
        description=(
            "If True, the correlation matrix is combined with an estimate based on the "
            "standardized returns. The halflife_factor_vra is used for the weights."
        ),
    )
    nw_lags_factor_vol: int = Field(
        0, description="The Newey-West lags of the factor volatility."
    )
    nw_lags_factor_vol_halflife_override: int | None = Field(
        None,
        description=(
            "The Newey-West lags override for the factor volatility for lagged returns."
        ),
    )
    nw_lags_factor_cor: int = Field(
        0, description="The Newey-West lags of the factor correlation."
    )
    nw_lags_idio_vol: int = Field(
        0, description="The Newey-West lags of the idiosyncratic volatility."
    )
    nw_lags_idio_vol_halflife_override: int | None = Field(
        None,
        description=(
            "The Newey-West lags override for the idiosyncratic volatility for lagged returns."
        ),
    )
    shrink_idio_vol: float = Field(
        0.0, description="The shrinkage factor for the idiosyncratic volatility."
    )
    clip_idio_return_lb: float | None = Field(
        -0.3,
        description=(
            "The lower bound to clip the idiosyncratic return in order to compute "
            "the idiosyncratic volatility. If None, no clipping is applied."
        ),
    )
    clip_idio_return_ub: float | None = Field(
        None,
        description=(
            "The upper bound to clip the idiosyncratic return in order to compute "
            "the idiosyncratic volatility. If None, no clipping is applied."
        ),
    )
    void_idio_return_lb: float | None = Field(
        None,
        description=(
            "The lower bound to remove the idiosyncratic return in order to compute "
            "the idiosyncratic volatility. Removed returns are treated as if there "
            "was a holiday for this asset. If None, no removal is applied. Voiding "
            "returns is applied before clipping."
        ),
    )
    void_idio_return_ub: float | None = Field(
        0.5,
        description=(
            "The upper bound to remove the idiosyncratic return in order to compute "
            "the idiosyncratic volatility. Removed returns are treated as if there "
            "was a holiday for this asset. If None, no removal is applied. Voiding "
            "returns is applied before clipping."
        ),
    )
    structural_model_fallback_halflife: int = Field(
        21,
        description=(
            "The half-life to use for the structural model fallback. The structural "
            "model computes idio vol based on the factor exposures. The halflife "
            "determines the dacay of this estimate when historical returns become "
            "available."
        ),
    )

    @property
    def menu_type(self) -> type[PortfolioStressTestReportSettingsMenu]:  # noqa: D102
        return PortfolioStressTestReportSettingsMenu


class AssetStressTestReportSettings(ConcreteReportSettings):
    """Defines settings to build an asset stress test report."""

    report_type: Literal["Asset Stress Test report"] = Field(
        "Asset Stress Test report", repr=False
    )
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(
            ALL_ASSET_STRESS_TEST_MEASURE_TYPES
        ),
        description="The measures to include in the report.",
    )
    halflife_factor_vol: int = Field(
        42, description="The half-life for the factor volatility."
    )
    halflife_factor_vra: int | None = Field(
        None,
        description=(
            "The half-life for the factor volatility regime adjustment. "
            "If None, no adjustment is applied."
        ),
    )
    halflife_factor_cor: int = Field(
        126, description="The half-life for the factor correlation."
    )
    halflife_idio_vol: int = Field(
        42, description="The half-life for the idiosyncratic volatility."
    )
    shrink_factor_cor_method: Literal["LIS", "QIS", "GIS"] | None = Field(
        None,
        description=(
            "The method to use for the shrinkage of the factor correlation. If None, "
            "don't shrink."
        ),
    )
    shrink_factor_cor_length: int | None = Field(
        1008,
        description=(
            "If a shrinkage method is provided, the length of the shrinkage for the "
            "factor correlation."
        ),
    )
    shrink_factor_cor_standardized: bool = Field(
        False,
        description=(
            "If True, the correlation matrix is combined with an estimate based on the "
            "standardized returns. The halflife_factor_vra is used for the weights."
        ),
    )
    nw_lags_factor_vol: int = Field(
        0, description="The Newey-West lags of the factor volatility."
    )
    nw_lags_factor_vol_halflife_override: int | None = Field(
        None,
        description=(
            "The Newey-West lags override for the factor volatility for lagged returns."
        ),
    )
    nw_lags_factor_cor: int = Field(
        0, description="The Newey-West lags of the factor correlation."
    )
    nw_lags_idio_vol: int = Field(
        0, description="The Newey-West lags of the idiosyncratic volatility."
    )
    nw_lags_idio_vol_halflife_override: int | None = Field(
        None,
        description=(
            "The Newey-West lags override for the idiosyncratic volatility for lagged returns."
        ),
    )
    shrink_idio_vol: float = Field(
        0.0, description="The shrinkage factor for the idiosyncratic volatility."
    )
    clip_idio_return_lb: float | None = Field(
        -0.3,
        description=(
            "The lower bound to clip the idiosyncratic return in order to compute "
            "the idiosyncratic volatility. If None, no clipping is applied."
        ),
    )
    clip_idio_return_ub: float | None = Field(
        None,
        description=(
            "The upper bound to clip the idiosyncratic return in order to compute "
            "the idiosyncratic volatility. If None, no clipping is applied."
        ),
    )
    void_idio_return_lb: float | None = Field(
        None,
        description=(
            "The lower bound to remove the idiosyncratic return in order to compute "
            "the idiosyncratic volatility. Removed returns are treated as if there "
            "was a holiday for this asset. If None, no removal is applied. Voiding "
            "returns is applied before clipping."
        ),
    )
    void_idio_return_ub: float | None = Field(
        0.5,
        description=(
            "The upper bound to remove the idiosyncratic return in order to compute "
            "the idiosyncratic volatility. Removed returns are treated as if there "
            "was a holiday for this asset. If None, no removal is applied. Voiding "
            "returns is applied before clipping."
        ),
    )
    structural_model_fallback_halflife: int = Field(
        21,
        description=(
            "The half-life to use for the structural model fallback. The structural "
            "model computes idio vol based on the factor exposures. The halflife "
            "determines the dacay of this estimate when historical returns become "
            "available."
        ),
    )

    @property
    def menu_type(self) -> type[AssetStressTestReportSettingsMenu]:  # noqa: D102
        return AssetStressTestReportSettingsMenu


class AssetHoldingsReportSettings(ConcreteReportSettings):
    """Defines settings to build an asset holdings report."""

    report_type: Literal["Asset Holdings report"] = Field(
        "Asset Holdings report", repr=False
    )
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(ALL_ASSET_HOLDINGS_MEASURE_TYPES),
        description="The measures to include in the report.",
    )

    @property
    def menu_type(self) -> type[AssetHoldingsReportSettingsMenu]:  # noqa: D102
        return AssetHoldingsReportSettingsMenu


ConcreteReportSettingsType: TypeAlias = Union[tuple(ConcreteReportSettings.__subclasses__())]  # type: ignore



import abc

from bayesline.api._src._utils import docstrings_from_sync
from bayesline.api._src.equity.modelconstruction_settings import (
    ModelConstructionSettings,
    ModelConstructionSettingsMenu,
)
from bayesline.api._src.registry import AsyncRegistryBasedApi, RegistryBasedApi


class FactorModelConstructionApi(abc.ABC):
    """Abstract base class for factor model construction APIs.

    This class defines the interface for factor model construction APIs that provide
    access to model construction settings and functionality.
    """

    @property
    @abc.abstractmethod
    def settings(self) -> ModelConstructionSettings:
        """Get the model construction settings.

        Returns
        -------
        ModelConstructionSettings
            The model construction settings.
        """
        ...


@docstrings_from_sync
class AsyncFactorModelConstructionApi(abc.ABC):

    @property
    @abc.abstractmethod
    def settings(self) -> ModelConstructionSettings: ...  # noqa: D102


class FactorModelConstructionLoaderApi(
    RegistryBasedApi[
        ModelConstructionSettings,
        ModelConstructionSettingsMenu,
        FactorModelConstructionApi,
    ],
):
    """API for loading factor model construction services.

    This class provides functionality for loading factor model construction services using
    registry-based API patterns.
    """

    ...


@docstrings_from_sync
class AsyncFactorModelConstructionLoaderApi(
    AsyncRegistryBasedApi[
        ModelConstructionSettings,
        ModelConstructionSettingsMenu,
        AsyncFactorModelConstructionApi,
    ],
): ...



import inspect
import re
import sys
from typing import Any, Callable, TypeVar

T = TypeVar("T", bound=type)

# ---------------- helpers ----------------


def _short_type(ann: Any) -> str:
    if ann is None:
        return "None"
    s = getattr(ann, "__name__", None) or str(ann)
    s = s.replace("typing.", "")
    return re.sub(r"(?<!\w)[\w\.]*\.", "", s)  # strip module prefixes


def _param_names(func: Callable[..., Any]) -> list[str]:
    sig = inspect.signature(func)
    return [p.name for p in sig.parameters.values() if p.name not in {"self", "cls"}]


def _find_numpy_block(
    doc: str, title: str
) -> tuple[tuple[int, int], tuple[int, int]] | None:
    pat = re.compile(rf"(^\s*{re.escape(title)}\s*\n^\s*-+\s*\n)", re.MULTILINE)
    m = pat.search(doc)
    if not m:
        return None
    body_start = m.end(1)
    body_m = re.compile(r"(?=^\s*\n^\S)|\Z", re.MULTILINE).search(doc, body_start)
    body_end = body_m.start() if body_m else len(doc)
    return (m.span(1), (body_start, body_end))


def _split_type_and_tail(r: str) -> tuple[str, str]:
    # split 'TYPE[, tail...]' at first comma not inside (), [], {}.
    depth = 0
    for i, ch in enumerate(r):
        if ch in "([{":
            depth += 1
        elif ch in ")]}":
            depth -= 1
        elif ch == "," and depth == 0:
            return r[:i].strip(), r[i:]  # keep comma in tail
    return r.strip(), ""


def _replace_idents_to_async(type_text: str, dst_ann_str: str) -> str:
    # replace only identifiers X with AsyncX if AsyncX appears in dst_ann_str.
    # - preserves generics/Literal contents and anything not explicitly mapped.
    # - only replaces CamelCase-like identifiers (starts with uppercase letter).

    def repl(m: re.Match) -> str:
        ident = m.group(0)
        if not ident[0].isupper():
            return ident
        cand = f"Async{ident}"
        return cand if re.search(rf"\b{re.escape(cand)}\b", dst_ann_str) else ident

    # Replace dotted names token-by-token (only the trailing identifier will match)
    return re.sub(r"\b[A-Za-z_][A-Za-z0-9_]*\b", repl, type_text)


def _rewrite_params_types_numpy(body: str, dst_ann_str_by_name: dict[str, str]) -> str:
    out = []
    for line in body.splitlines():
        m = re.match(r"^(\s*)([A-Za-z_]\w*)\s*:\s*(.+)$", line)
        if not m:
            out.append(line)
            continue
        indent, name, right = m.groups()
        if name not in dst_ann_str_by_name:
            out.append(line)
            continue
        type_token, tail = _split_type_and_tail(right.lstrip())
        # Only tweak identifiers inside the existing type token
        new_type = _replace_idents_to_async(type_token, dst_ann_str_by_name[name])
        out.append(f"{indent}{name} : {new_type}{tail}")
    return "\n".join(out)


def _rewrite_returns_type_numpy(body: str, dst_ret_str: str) -> str:
    lines = body.splitlines()
    for i, line in enumerate(lines):
        s = line.strip()
        if s and not set(s) <= {"-"}:
            indent = line[: len(line) - len(line.lstrip())]
            keep_colon = ":" if s.endswith(":") else ""
            new_type = _replace_idents_to_async(s.rstrip(":"), dst_ret_str)
            lines[i] = f"{indent}{new_type}{keep_colon}"
            break
    return "\n".join(lines)


def _adapt_doc_numpy(
    src_doc: str, src_func: Callable[..., Any], dst_func: Callable[..., Any]
) -> str:
    new_doc = src_doc

    # If param names match, adapt only the identifiers inside the existing type text
    if _param_names(src_func) == _param_names(dst_func):
        dst_sig = inspect.signature(dst_func)
        dst_ann_str_by_name = {
            n: _short_type(dst_func.__annotations__.get(n, None))
            for n in dst_sig.parameters
            if n not in {"self", "cls"}
        }
        blk = _find_numpy_block(new_doc, "Parameters")
        if blk:
            (_, _), (bs, be) = blk
            new_doc = (
                new_doc[:bs]
                + _rewrite_params_types_numpy(new_doc[bs:be], dst_ann_str_by_name)
                + new_doc[be:]
            )

    # Returns: conservatively replace identifiers only
    dst_ret_str = _short_type(dst_func.__annotations__.get("return"))
    blk = _find_numpy_block(new_doc, "Returns")
    if blk and dst_ret_str:
        (_, _), (bs, be) = blk
        new_doc = (
            new_doc[:bs]
            + _rewrite_returns_type_numpy(new_doc[bs:be], dst_ret_str)
            + new_doc[be:]
        )

    return new_doc


# ---------------- main decorator ----------------


def docstrings_from_sync(cls: T) -> T:  # noqa: C901
    """For an async twin class (e.g., AsyncFoo), copy/adjust NumPy docstrings from Foo.

    - Copy class docstring if AsyncFoo lacks one.
    - For each method *and property* in AsyncFoo:
        * If it lacks a docstring, copy Foo counterpart's docstring.
        * Adapt only identifier tokens (X -> AsyncX) inside 'Parameters'/'Returns' types.
          Do not expand generics or Literal values.
        * For properties, set docstring on both the property object and its fget.

    Parameters
    ----------
    cls : T
        The async twin class to copy docstrings into.

    Returns
    -------
    T
        The async twin class with copied docstrings.
    """
    name = cls.__name__
    sync_name = name[5:] if name.startswith("Async") else name.replace("Async", "", 1)
    mod = sys.modules.get(cls.__module__)
    sync_cls = getattr(mod, sync_name, None) if mod else None
    if not isinstance(sync_cls, type):
        return cls

    # Class docstring
    if not cls.__doc__ and getattr(sync_cls, "__doc__", None):
        cls.__doc__ = sync_cls.__doc__

    for attr, dst in list(cls.__dict__.items()):
        if attr.startswith("_"):
            continue

        # properties
        if isinstance(dst, property):
            src_prop = getattr(sync_cls, attr, None)
            if not isinstance(src_prop, property):
                continue
            src_doc = src_prop.__doc__
            if not src_doc:
                continue
            dst_fget = dst.fget
            src_fget = src_prop.fget
            if not (dst_fget and src_fget):
                continue
            new_doc = _adapt_doc_numpy(src_doc, src_fget, dst_fget)
            dst_fget.__doc__ = new_doc
            setattr(cls, attr, property(dst.fget, dst.fset, dst.fdel, doc=new_doc))
            continue

        # functions (sync or async)
        if inspect.isfunction(dst):
            src = getattr(sync_cls, attr, None)
            if not (src and inspect.isfunction(src)):
                continue
            if getattr(dst, "__doc__", None):
                continue
            src_doc = getattr(src, "__doc__", None)
            if not src_doc:
                continue
            new_doc = _adapt_doc_numpy(src_doc, src, dst)
            try:
                dst.__doc__ = new_doc
            except Exception:
                dst.__func__.__doc__ = new_doc  # type: ignore[attr-defined]

    return cls



from __future__ import annotations

import json
import os
from collections.abc import Mapping, Sequence
from typing import Annotated, Any, Literal

import polars as pl
from pydantic import (
    BaseModel,
    Field,
    NonNegativeFloat,
    ValidationInfo,
    field_validator,
    model_validator,
)

from bayesline.api._src.equity import settings as settings_tools
from bayesline.api._src.equity.calendar_settings import (
    CalendarSettings,
    CalendarSettingsMenu,
)
from bayesline.api._src.equity.portfolio_settings import PortfolioOrganizerSettings
from bayesline.api._src.registry import Settings, SettingsMenu, SettingsTypeMetaData
from bayesline.api._src.types import IdType

Hierarchy = settings_tools.Hierarchy


class CategoricalFilterSettings(BaseModel, frozen=True, extra="forbid"):
    """Specify include and exclude filters for categorical codes.

    Examples of categorical codes are industries or countries. Assets are included if
    they are part of at least one include and not part of any exclude.

    By default all codes for the given hierarchy are included.
    """

    hierarchy: str = Field(
        min_length=1,
        description="The categorical hierarchy to use.",
        examples=["trbc"],
    )

    include: list[str] | Literal["All"] = Field(
        default="All",
        description=(
            "Valid industry codes or labels for given hierarchy at any level. If "
            "labels are used which may be duplicated, then the code with the highest "
            "level is used. If 'All', all codes are included."
        ),
        examples=[["3571"], "All", ["Materials", "1010"], ["Europe", "CAN"]],
    )

    exclude: list[str] = Field(
        default_factory=list,
        description=(
            "Valid industry codes or labels for given hierarchy at any level. If "
            "labels are used which may be duplicated, then the code with the lowest "
            "level is used."
        ),
        examples=[["3571"], ["Materials", "1010"], ["JPN"]],
    )

    @model_validator(mode="after")
    def validate_filters(self, info: ValidationInfo) -> CategoricalFilterSettings:
        if isinstance(menu := info.context, UniverseSettingsMenu):
            settings_tools.validate_hierarchy_schema(
                menu.categorical_hierarchies, self.hierarchy
            )
            settings_tools.validate_hierarchy_filters(
                menu.categorical_hierarchies[self.hierarchy],
                self.include,
                self.exclude,
                menu.categorical_hierarchies_labels[self.hierarchy],
            )
        return self

    def validate_settings(self, menu: UniverseSettingsMenu | None) -> None:
        """Validate the categorical filter settings.

        Parameters
        ----------
        menu : UniverseSettingsMenu | None
            The menu to validate against. If None, the categorical filter settings is validated
            without having the menu context. This means that less validation is done.
        """
        self.model_validate(self.model_dump(), context=menu)


class MCapFilterSettings(BaseModel, frozen=True, extra="forbid"):
    """Specify the lower and upper bound for the market cap filter.

    By default the bounds are infinite.
    """

    lower: NonNegativeFloat = Field(
        default=0.0,
        ge=0.0,
        description="Lower bound of the cap filter in USD.",
        examples=[1e10],
    )

    upper: NonNegativeFloat = Field(
        default=1e20,
        gt=0.0,
        description="Upper bound of the cap filter in USD.",
        examples=[1e12],
    )

    @model_validator(mode="after")
    def check_upper_gt_lower(self) -> MCapFilterSettings:
        """Validate that the upper bound is greater than the lower bound.

        Returns
        -------
        MCapFilterSettings
            The validated instance.

        Raises
        ------
        ValueError
            If the upper bound is not greater than the lower bound.
        """
        if (lower := self.lower) >= (upper := self.upper):
            raise ValueError(
                f"upper bound {upper} must be greater than lower bound {lower}",
            )
        else:
            return self


class UniverseSettingsMenu(SettingsMenu, frozen=True, extra="forbid"):
    """Contain the available settings that can be used for the universe settings."""

    calendar_settings_menu: CalendarSettingsMenu = Field(
        description="The calendar settings menu to use for the universe.",
    )

    id_types: list[IdType] = Field(
        description="""
        A list of all the id types that are supported for the universe.
        """,
    )

    categorical_hierarchies: Mapping[str, Hierarchy] = Field(
        description="""
        A dictionary where the key is the name of the categorical hierarchy (e.g. 'trbc')
        and the value is a N-level nested dictionary structure of the categorical hierarchy
        codes.
        """,
    )

    categorical_hierarchies_labels: Mapping[str, Mapping[str, str]] = Field(
        description="""
        A dictionary where the key is the name of the categorical hierarchy and
        the value is a mapping from unique categorical code to a human readable name.
        """,
    )

    @field_validator("categorical_hierarchies")
    @classmethod
    def check_unique_hierarchy(
        cls: type[UniverseSettingsMenu], v: Mapping[str, Hierarchy]
    ) -> Mapping[str, Hierarchy]:
        return settings_tools.check_unique_hierarchy(v)

    @field_validator("categorical_hierarchies")
    @classmethod
    def check_no_empty_branches(
        cls,
        v: Mapping[str, Mapping[str, Hierarchy]],
    ) -> Mapping[str, Mapping[str, Hierarchy]]:
        for hierarchy_name, hierarchy in v.items():
            if not hierarchy:  # different error message at root level
                raise ValueError(f"Hierarchy '{hierarchy_name}' cannot be empty")
            settings_tools.assert_no_empty_branches(hierarchy, hierarchy_name)
        return v

    @model_validator(mode="after")
    def check_all_codes_have_labels(self) -> UniverseSettingsMenu:
        if errors := settings_tools.check_all_codes_have_labels(
            self.categorical_hierarchies,
            self.categorical_hierarchies_labels,
        ):
            raise ValueError(os.linesep.join(errors))
        else:
            return self

    def describe(self) -> str:
        """Generate a human-readable description of the universe settings.

        Returns
        -------
        str
            A formatted description of the universe settings menu.
        """
        hierarchies = self.categorical_hierarchies
        labels = self.categorical_hierarchies_labels
        hierarchies_str = json.dumps(
            {
                k: settings_tools.codes_to_labels(hierarchies[k], labels[k])
                for k in hierarchies
            },
            indent=2,
        )
        id_types = ", ".join(self.id_types)
        description = [
            f"ID Types: {id_types}",
            "Hierarchies:",
            hierarchies_str,
        ]

        return os.linesep.join(description)

    def effective_categories(
        self,
        settings: CategoricalFilterSettings,
        labels: bool = False,
    ) -> list[str]:
        """Get the effective leaf level categorical codes after categorical filtering.

        Parameters
        ----------
        settings : CategoricalFilterSettings
            The filter settings to get the effective categorical codes for.
        labels : bool, default=False
            Whether to return the labels or the codes.

        Returns
        -------
        list[str]
            The effective leaf level categorical codes for the given settings after the
            filters were applied.
        """
        settings.validate_settings(self)
        effective_codes = settings_tools.effective_leaves(
            self.categorical_hierarchies[settings.hierarchy],
            settings.include,
            settings.exclude,
            self.categorical_hierarchies_labels[settings.hierarchy],
        )
        if labels:
            return [
                self.categorical_hierarchies_labels[settings.hierarchy][code]
                for code in effective_codes
            ]
        else:
            return effective_codes

    def hierarchy_df(self, hierarchy: str) -> pl.DataFrame:
        """Return a dataframe of the given categorical hierarchy.

        Parameters
        ----------
        hierarchy : str
            The name of the categorical hierarchy to return.

        Returns
        -------
        pl.DataFrame
            The wide DataFrame representation of the hierarchy. The columns are:
            - level_1: The code of the root level.
            - level_1_label: The label of the root level.
            - level_2: The code of the second level.
            - level_2_label: The label of the second level.
            - ...
            - level_n: The code of the n-th level.
            - level_n_label: The label of the n-th level.
        """
        return settings_tools.hierarchy_df_to_wide(
            self.categorical_hierarchies[hierarchy],
            self.categorical_hierarchies_labels[hierarchy],
        )


class UniverseSettings(Settings[UniverseSettingsMenu]):
    """Define an asset universe as a set of regional, industry and market cap filters."""

    dataset: str = Field(
        description=(
            "The name of the underlying dataset to use. If none is given then the "
            "configured default dataset is used."
        ),
        examples=["Bayesline-Global"],
    )

    id_type: IdType = Field(
        default="bayesid",
        description="The default id type to use for the universe.",
        examples=["cusip9", "bayesid"],
    )

    calendar: CalendarSettings = Field(
        default_factory=CalendarSettings,
        description="The calendar settings to use for the universe.",
    )

    categorical_filters: list[CategoricalFilterSettings] = Field(
        default_factory=list,
        description="""
        Filters that determine which categorical codes to include and exclude in the universe.
        """,
    )

    portfolio_filter: Annotated[
        str | int | PortfolioOrganizerSettings | None,
        Field(
            description=(
                "The portfolio organizer settings to use as an underlying schema of "
                "portfolios. Universe will be filtered across the superset of all "
                "portfolios."
            ),
            default=None,
        ),
        SettingsTypeMetaData[str | int | PortfolioOrganizerSettings](
            references=PortfolioOrganizerSettings
        ),
    ] = None

    mcap_filter: MCapFilterSettings = Field(
        default_factory=MCapFilterSettings,
        description="""
        Filters that determine which market caps to include and exclude in the universe.
        """,
    )

    def get_references(self) -> Sequence[str | int]:
        """
        Get references for this universe settings object.

        Returns
        -------
        Sequence[str | int]
            A sequence of references (strings or integers) for this universe
            settings object.
        """
        references: list[str | int] = [self.dataset]

        if self.portfolio_filter is not None:
            if isinstance(self.portfolio_filter, PortfolioOrganizerSettings):
                references.extend(self.portfolio_filter.get_references())
            else:
                references.append(self.portfolio_filter)

        return references

    @property
    def menu_type(self) -> type[UniverseSettingsMenu]:  # noqa: D102
        return UniverseSettingsMenu

    @field_validator("categorical_filters")
    @classmethod
    def check_unique_filters(
        cls: type[UniverseSettings], v: list[CategoricalFilterSettings]
    ) -> list[CategoricalFilterSettings]:
        factor_groups = [filter_settings.hierarchy for filter_settings in v]
        if len(factor_groups) != len(set(factor_groups)):
            raise ValueError("categorical_filters must reference unique hierarchies")
        return v

    @model_validator(mode="before")
    @classmethod
    def propagate_dataset(cls: type[UniverseSettings], data: Any) -> Any:
        if isinstance(data, dict) and "dataset" in data:
            if "calendar" not in data:
                data["calendar"] = {}
            if isinstance(data["calendar"], dict):
                if data["calendar"].get("dataset") is None:
                    data["calendar"]["dataset"] = data["dataset"]
        return data

    @field_validator("id_type", mode="after")
    @classmethod
    def validate_id_type(cls, id_type: IdType, info: ValidationInfo) -> IdType:
        if isinstance(menu := info.context, UniverseSettingsMenu):
            if id_type not in menu.id_types:
                raise ValueError(
                    f"Id type {id_type} does not exist. Only "
                    f"{', '.join(menu.id_types)} exist."
                )
        return id_type

    @field_validator("calendar", mode="after")
    @classmethod
    def validate_calendar(
        cls, calendar_settings: CalendarSettings, info: ValidationInfo
    ) -> CalendarSettings:
        if isinstance(menu := info.context, UniverseSettingsMenu):
            calendar_settings.validate_settings(menu.calendar_settings_menu)
        return calendar_settings

    @field_validator("categorical_filters", mode="after")
    @classmethod
    def validate_categorical_filters(
        cls, filter_settings: list[CategoricalFilterSettings], info: ValidationInfo
    ) -> list[CategoricalFilterSettings]:
        if isinstance(menu := info.context, UniverseSettingsMenu):
            for filter_setting in filter_settings:
                filter_setting.model_validate(filter_setting.model_dump(), context=menu)
        return filter_settings

    @field_validator("portfolio_filter", mode="after")
    @classmethod
    def validate_portfolio_filter(
        cls,
        portfolio_filter: str | int | PortfolioOrganizerSettings | None,
        info: ValidationInfo,
    ) -> str | int | PortfolioOrganizerSettings | None:
        if isinstance(portfolio_filter, PortfolioOrganizerSettings):
            dataset = info.data["dataset"]
            if portfolio_filter.dataset is None:
                portfolio_filter = portfolio_filter.model_copy(
                    update={"dataset": dataset}
                )
            if portfolio_filter.dataset != dataset:
                raise ValueError(
                    f"The dataset in the portfolio filter, {portfolio_filter.dataset} "
                    f"must match the dataset, {dataset}."
                )
        return portfolio_filter

    def describe(self, menu: UniverseSettingsMenu) -> str:
        """Describe the universe settings.

        Parameters
        ----------
        menu : UniverseSettingsMenu
            The settings menu to use to describe the settings.

        Returns
        -------
        str
            A description of the universe settings.
        """
        hierarchies = menu.categorical_hierarchies
        labels = menu.categorical_hierarchies_labels

        self.validate_settings(menu)  # TODO: needed?
        description = [f"Default ID Type: {self.id_type!r}"]

        for filter_settings in self.categorical_filters:
            hierarchy_name = filter_settings.hierarchy
            effective_hierarchy = settings_tools.effective_hierarchy(
                hierarchies[hierarchy_name],
                filter_settings.include,
                filter_settings.exclude,
                labels[hierarchy_name],
            )
            effective_hierarchy = settings_tools.codes_to_labels(
                effective_hierarchy, labels[hierarchy_name]
            )
            description.extend(
                [
                    f"Hierarchy ({hierarchy_name}):",
                    json.dumps(effective_hierarchy, indent=2),
                ]
            )

        description.extend(
            [
                "Market Cap:",
                self.mcap_filter.model_dump_json(indent=2),
            ]
        )
        return os.linesep.join(description)



import abc
import datetime as dt
from typing import Literal, overload

import polars as pl

from bayesline.api._src._utils import docstrings_from_sync
from bayesline.api._src.equity.riskmodels_settings import (
    FactorRiskModelMetadata,
    FactorRiskModelSettings,
    FactorRiskModelSettingsMenu,
    GetModelMode,
)
from bayesline.api._src.registry import AsyncRegistryBasedApi, RegistryBasedApi
from bayesline.api._src.tasks import AsyncTask, Task
from bayesline.api._src.types import DateLike, IdType


class FactorModelApi(abc.ABC):
    """Provide access to factor risk model data and operations.

    A factor risk model provides exposure data, factor returns, and other risk metrics
    for a set of assets based on predefined factors.
    """

    @abc.abstractmethod
    def dates(self) -> list[dt.date]:
        """Get all dates covered by this risk model.

        Returns
        -------
        list[dt.date]
            All dates covered by this risk model.
        """
        pass

    @abc.abstractmethod
    def factors(self) -> dict[str, list[str]]:
        """Get the factor groups and their factors.

        Returns
        -------
        dict[str, list[str]]
            Dict where the keys are the included factor groups, and the values are the
            factors in that group.
        """
        ...

    @abc.abstractmethod
    def exposures(
        self,
        *,
        start: DateLike | None = None,
        end: DateLike | None = None,
        id_type: IdType | None = None,
        stage: int = 1,
    ) -> pl.DataFrame:
        """
        Get the risk model exposures for this risk model.

        Parameters
        ----------
        start : DateLike | None, default=None
            The start date of the data to return, inclusive.
        end : DateLike | None, default=None
            The end date of the data to return, inclusive.
        id_type : IdType | None, default=None
            The id type to return asset ids in, e.g. `ticker`.
            The given id type must be supported by the linked universe.
        stage : int, default=1
            The stage of the factor model to return exposures for, of the potentially
            multi-stage regression. Default to 1 (the first stage).

        Raises
        ------
        ValueError
            If the given id type is not supported or date range is invalid.

        Returns
        -------
        pl.DataFrame
            The data for the given date range with the first two column as the date and
            asset id. The remaining columns are the individual styles.
        """
        ...

    @abc.abstractmethod
    def universe(
        self,
        *,
        start: DateLike | None = None,
        end: DateLike | None = None,
        id_type: IdType | None = None,
        stage: int = 1,
    ) -> pl.DataFrame:
        """
        Get the risk model universe for this risk model.

        Parameters
        ----------
        start : DateLike | None, default=None
            The start date of the data to return, inclusive.
        end : DateLike | None, default=None
            The end date of the data to return, inclusive.
        id_type : IdType | None, default=None
            The id type to return asset ids in, e.g. `ticker`.
            The given id type must be supported by the linked universe.
        stage : int, default=1
            The stage of the factor model to return universe for, of the potentially
            multi-stage regression. Default to 1 (the first stage).

        Raises
        ------
        ValueError
            If the given id type is not supported or date range is invalid.

        Returns
        -------
        pl.DataFrame
            The data for the given date range where the first column is the date and the
            remaining columns are the asset ids. The values are the universe inclusion.
        """
        ...

    @abc.abstractmethod
    def estimation_universe(
        self,
        *,
        start: DateLike | None = None,
        end: DateLike | None = None,
        id_type: IdType | None = None,
        stage: int = 1,
    ) -> pl.DataFrame:
        """
        Get the risk model estimation universe for this risk model.

        Parameters
        ----------
        start : DateLike | None, default=None
            The start date of the data to return, inclusive.
        end : DateLike | None, default=None
            The end date of the data to return, inclusive.
        id_type : IdType | None, default=None
            The id type to return asset ids in, e.g. `ticker`.
            The given id type must be supported by the linked universe.
        stage : int, default=1
            The stage of the factor model to return universe for, of the potentially
            multi-stage regression. Default to 1 (the first stage).

        Raises
        ------
        ValueError
            If the given id type is not supported or date range is invalid.

        Returns
        -------
        pl.DataFrame
            The data for the given date range where the first column is the date and the
            remaining columns are the asset ids. The values are the estimation universe
            inclusion.
        """
        ...

    @abc.abstractmethod
    def market_caps(
        self,
        *,
        start: DateLike | None = None,
        end: DateLike | None = None,
        id_type: IdType | None = None,
    ) -> pl.DataFrame:
        """
        Get the exposure-weighted market caps for this risk model.

        Parameters
        ----------
        start : DateLike | None, default=None
            The start date of the data to return, inclusive.
        end : DateLike | None, default=None
            The end date of the data to return, inclusive.
        id_type : IdType | None, default=None
            The id type to return asset ids in, e.g. `ticker`.
            The given id type must be supported by the linked universe.

        Raises
        ------
        ValueError
            If the given id type is not supported or date range is invalid.

        Returns
        -------
        pl.DataFrame
            The data for the given date range where the rows are the total market cap
            for each date, weighted by the exposure of each asset. For industry factors,
            this specifically means that the value is the sum of all assets in the
            estimation universe in that industry.
        """
        ...

    @abc.abstractmethod
    def weights(
        self,
        *,
        start: DateLike | None = None,
        end: DateLike | None = None,
        id_type: IdType | None = None,
    ) -> pl.DataFrame:
        """
        Get the idiosyncratic volatility weights for this risk model.

        Parameters
        ----------
        start : DateLike | None, default=None
            The start date of the data to return, inclusive.
        end : DateLike | None, default=None
            The end date of the data to return, inclusive.
        id_type : IdType | None, default=None
            The id type to return asset ids in, e.g. `ticker`.
            The given id type must be supported by the linked universe.

        Raises
        ------
        ValueError
            If the given id type is not supported or date range is invalid.

        Returns
        -------
        pl.DataFrame
            The data for the given date range where the rows are the idiosyncratic
            volatility for each date.
        """
        ...

    @abc.abstractmethod
    def future_asset_returns(
        self,
        *,
        start: DateLike | None = None,
        end: DateLike | None = None,
        id_type: IdType | None = None,
    ) -> pl.DataFrame:
        """
        Get the asset returns for this risk model on the next day.

        Parameters
        ----------
        start : DateLike | None, default=None
            The start date of the data to return, inclusive.
        end : DateLike | None, default=None
            The end date of the data to return, inclusive.
        id_type : IdType | None, default=None
            The id type to return asset ids in, e.g. `ticker`.
            The given id type must be supported by the linked universe.

        Raises
        ------
        ValueError
            If the given id type is not supported or date range is invalid.

        Returns
        -------
        pl.DataFrame
            The data for the given date range where the index is the date
            and the columns are the asset id. The values are the asset returns.
        """
        ...

    @abc.abstractmethod
    def fret(
        self,
        *,
        freq: str | None = None,
        cumulative: bool = False,
        start: DateLike | None = None,
        end: DateLike | None = None,
    ) -> pl.DataFrame:
        """Get factor returns for the given date range.

        Parameters
        ----------
        freq : str | None, default=None
            The frequency of the return aggregation, e.g. `D` for daily.
            Defaults to daily (i.e. unaggregated).
        cumulative : bool, default=False
            If True, returns the cumulative returns.
        start : DateLike | None, default=None
            The start date of the data to return, inclusive.
        end : DateLike | None, default=None
            The end date of the data to return, inclusive.

        Returns
        -------
        pl.DataFrame
            The factor returns for the given date range.
        """
        ...

    @abc.abstractmethod
    def t_stats(self) -> pl.DataFrame:
        """Get t-statistics for the factor model.

        Returns
        -------
        pl.DataFrame
            The t-statistics for the factor model.
        """
        ...

    @abc.abstractmethod
    def p_values(self) -> pl.DataFrame:
        """Get p-values for the risk model factors.

        Returns
        -------
        pl.DataFrame
            The p-values for the risk model factors.
        """
        ...

    @abc.abstractmethod
    def r2(self) -> pl.DataFrame:
        """Get R-squared values for the factor model.

        Returns
        -------
        pl.DataFrame
            The R-squared values for the factor model.
        """
        ...

    @abc.abstractmethod
    def sigma2(self) -> pl.DataFrame:
        """Get sigma squared values for the factor model.

        Returns
        -------
        pl.DataFrame
            The sigma squared values for the factor model.
        """
        ...


@docstrings_from_sync
class AsyncFactorModelApi(abc.ABC):

    @abc.abstractmethod
    async def dates(self) -> list[dt.date]: ...  # noqa: D102

    @abc.abstractmethod
    async def factors(self) -> dict[str, list[str]]: ...  # noqa: D102

    @abc.abstractmethod
    async def exposures(  # noqa: D102
        self,
        *,
        start: DateLike | None = None,
        end: DateLike | None = None,
        id_type: IdType | None = None,
        stage: int = 1,
    ) -> pl.DataFrame: ...

    @abc.abstractmethod
    async def universe(  # noqa: D102
        self,
        *,
        start: DateLike | None = None,
        end: DateLike | None = None,
        id_type: IdType | None = None,
        stage: int = 1,
    ) -> pl.DataFrame: ...

    @abc.abstractmethod
    async def estimation_universe(  # noqa: D102
        self,
        *,
        start: DateLike | None = None,
        end: DateLike | None = None,
        id_type: IdType | None = None,
        stage: int = 1,
    ) -> pl.DataFrame: ...

    @abc.abstractmethod
    async def market_caps(  # noqa: D102
        self,
        *,
        start: DateLike | None = None,
        end: DateLike | None = None,
        id_type: IdType | None = None,
    ) -> pl.DataFrame: ...

    @abc.abstractmethod
    async def weights(  # noqa: D102
        self,
        *,
        start: DateLike | None = None,
        end: DateLike | None = None,
        id_type: IdType | None = None,
    ) -> pl.DataFrame: ...

    @abc.abstractmethod
    async def future_asset_returns(  # noqa: D102
        self,
        *,
        start: DateLike | None = None,
        end: DateLike | None = None,
        id_type: IdType | None = None,
    ) -> pl.DataFrame: ...

    @abc.abstractmethod
    async def fret(  # noqa: D102
        self,
        *,
        freq: str | None = None,
        cumulative: bool = False,
        start: DateLike | None = None,
        end: DateLike | None = None,
    ) -> pl.DataFrame: ...

    @abc.abstractmethod
    async def t_stats(self) -> pl.DataFrame: ...  # noqa: D102

    @abc.abstractmethod
    async def p_values(self) -> pl.DataFrame: ...  # noqa: D102

    @abc.abstractmethod
    async def r2(self) -> pl.DataFrame: ...  # noqa: D102

    @abc.abstractmethod
    async def sigma2(self) -> pl.DataFrame: ...  # noqa: D102


class FactorModelEngineApi(abc.ABC):
    """Provide access to factor risk model engine operations.

    A factor model engine is responsible for building and managing factor risk models
    based on provided settings.
    """

    @property
    @abc.abstractmethod
    def settings(self) -> FactorRiskModelSettings:
        """Get the settings used to create this risk model.

        Returns
        -------
        FactorRiskModelSettings
            The settings used to create this risk model.
        """
        ...

    @abc.abstractmethod
    def get_model(
        self, mode: GetModelMode = "get-or-compute-and-persist"
    ) -> FactorModelApi:
        """Get a built factor model instance for the given settings.

        Parameters
        ----------
        mode : GetModelMode, default="get-or-compute-and-persist"
            The mode to use to get the model.
            If `compute`, will compute the model from scratch and return it, without
            persisting it.
            If `get-or-compute-and-persist`, will check if the model is already
            persisted (and up to date) and return it if so, or else compute it
            from scratch and persist it.
            If `get-or-compute`, will check if the model is already
            persisted and return it if so, or else compute it from scratch and return
            it without persisting it.
            If `compute-and-persist`, will compute the model from scratch and persist it.
            If `get-or-fail`, will return the model if it is already persisted, or else
            raise an error if the model is not persisted.

        Returns
        -------
        FactorModelApi
            A built factor model instance for the given settings.
        """

    @abc.abstractmethod
    def get_model_as_task(
        self, mode: GetModelMode = "get-or-compute-and-persist"
    ) -> Task[FactorModelApi]:
        """Get a built factor model instance as a task.

        Parameters
        ----------
        mode : GetModelMode, default="get-or-compute-and-persist"
            The mode to use to get the model.
            If `compute`, will compute the model from scratch and return it, without
            persisting it.
            If `get-or-compute-and-persist`, will check if the model is already
            persisted (and up to date) and return it if so, or else compute it
            from scratch and persist it.
            If `get-or-compute`, will check if the model is already
            persisted and return it if so, or else compute it from scratch and return
            it without persisting it.
            If `compute-and-persist`, will compute the model from scratch and persist it.
            If `get-or-fail`, will return the model if it is already persisted, or else
            raise an error if the model is not persisted.

        Returns
        -------
        Task[FactorModelApi]
            A task that will return a built factor model instance
            for the given settings.
        """
        ...


@docstrings_from_sync
class AsyncFactorModelEngineApi(abc.ABC):

    @property
    @abc.abstractmethod
    def settings(self) -> FactorRiskModelSettings: ...  # noqa: D102

    @abc.abstractmethod
    async def get_model(  # noqa: D102
        self, mode: GetModelMode = "get-or-compute-and-persist"
    ) -> AsyncFactorModelApi: ...

    @abc.abstractmethod
    async def get_model_as_task(  # noqa: D102
        self, mode: GetModelMode = "get-or-compute-and-persist"
    ) -> AsyncTask[AsyncFactorModelApi]: ...


class FactorModelMetadataApi(abc.ABC):
    """Sub interface to list risk models in the registry."""

    @overload
    def list_riskmodels(
        self, risk_dataset: str | None = None, *, df: Literal[False] = ...
    ) -> list[FactorRiskModelMetadata]: ...

    @overload
    def list_riskmodels(
        self, risk_dataset: str | None = None, *, df: Literal[True] = ...
    ) -> pl.DataFrame: ...

    @abc.abstractmethod
    def list_riskmodels(
        self, risk_dataset: str | None = None, *, df: bool = False
    ) -> list[FactorRiskModelMetadata] | pl.DataFrame:
        """List all risk models in the registry.

        Parameters
        ----------
        risk_dataset : str | None, default=None
            The risk dataset to filter by.
            If not given, all risk models are returned.

        df : bool, default=False
            If True, return a polars DataFrame instead of a list of metadata objects.

        Returns
        -------
        list[FactorRiskModelMetadata] | pl.DataFrame
            The metadata of all risk models in the registry, either as a list of objects
            or as a DataFrame depending on the `df` parameter.
        """


@docstrings_from_sync
class AsyncFactorModelMetadataApi(abc.ABC):

    @overload
    async def list_riskmodels(
        self, risk_dataset: str | None = None, *, df: Literal[False] = ...
    ) -> list[FactorRiskModelMetadata]: ...

    @overload
    async def list_riskmodels(
        self, risk_dataset: str | None = None, *, df: Literal[True] = ...
    ) -> pl.DataFrame: ...

    @abc.abstractmethod
    async def list_riskmodels(  # noqa: D102
        self, risk_dataset: str | None = None, *, df: bool = False
    ) -> list[FactorRiskModelMetadata] | pl.DataFrame: ...


class FactorModelLoaderApi(
    RegistryBasedApi[
        FactorRiskModelSettings, FactorRiskModelSettingsMenu, FactorModelEngineApi
    ],
    FactorModelMetadataApi,
):
    """Provide access to factor risk model loading and registry operations.

    A factor model loader manages the registry of factor risk models and provides
    access to existing models.
    """


@docstrings_from_sync
class AsyncFactorModelLoaderApi(
    AsyncRegistryBasedApi[
        FactorRiskModelSettings, FactorRiskModelSettingsMenu, AsyncFactorModelEngineApi
    ],
    AsyncFactorModelMetadataApi,
):

    pass



import abc
from typing import Any, Literal

from pydantic import Field, field_validator

from bayesline.api._src._utils import docstrings_from_sync
from bayesline.api._src.equity.modelconstruction_settings import WeightingScheme
from bayesline.api._src.equity.portfolioreport import (
    AsyncReportAccessorApi,
    ReportAccessorApi,
)
from bayesline.api._src.equity.report.accessor import (
    AsyncTypedReportAccessorApi,
    TypedReportAccessorApi,
)
from bayesline.api._src.equity.report.api import AsyncReportApi, ReportApi
from bayesline.api._src.equity.report.settings import ReportSettings
from bayesline.api._src.equity.universe_settings import UniverseSettings
from bayesline.api._src.tasks import AsyncTask, Task
from bayesline.api._src.types import DateLike


class WeightsReportSettings(ReportSettings):
    """Settings for a weights report."""

    report_type: Literal["weights"] = Field(
        default="weights", description="Report type discriminator"
    )
    universe_settings: UniverseSettings = Field()
    weights: WeightingScheme = Field(
        description="The weighting scheme to use for the weights report.",
        default="InvIdioVar",
        examples=["InvIdioVar"],
    )

    @field_validator("weights", mode="after")
    @classmethod
    def check_valid_weights(cls, weights: WeightingScheme) -> WeightingScheme:
        if weights != "InvIdioVar":
            raise ValueError(f"Weights {weights} are not supported for weights report.")
        return weights


class WeightsReportAccessor(TypedReportAccessorApi):
    """Accessor for a weights report."""

    pass


@docstrings_from_sync
class AsyncWeightsReportAccessor(AsyncTypedReportAccessorApi):
    pass


class WeightsReportApi(
    ReportApi[
        [DateLike | None, DateLike | None],
        WeightsReportAccessor,
        WeightsReportSettings,
    ]
):
    """API for a weights report."""

    @abc.abstractmethod
    def get_maybe_cached(
        self,
        start_date: DateLike | None,
        end_date: DateLike | None,
        **kwargs: Any,
    ) -> WeightsReportAccessor | None:
        """
        Get the cached weights report accessor API if it is available.

        Parameters
        ----------
        start_date: DateLike | None
            The start date of the report.
        end_date: DateLike | None
            The end date of the report.
        **kwargs: Any
            Additional keyword arguments.

        Returns
        -------
        WeightsReportAccessor | None
            The cached weights report accessor API, or None if not cached.
        """

    @abc.abstractmethod
    def calculate(
        self,
        start_date: DateLike | None,
        end_date: DateLike | None,
        **kwargs: Any,
    ) -> WeightsReportAccessor:
        """
        Calculate the weights report.

        Parameters
        ----------
        start_date: DateLike | None
            The start date of the report.
        end_date: DateLike | None
            The end date of the report.
        **kwargs: Any
            Additional keyword arguments.

        Returns
        -------
        WeightsReportAccessor
            The weights report accessor.
        """

    @abc.abstractmethod
    def calculate_as_task(
        self,
        start_date: DateLike | None,
        end_date: DateLike | None,
        **kwargs: Any,
    ) -> Task[WeightsReportAccessor]:
        """
        Calculate the weights report as a task.

        Parameters
        ----------
        start_date: DateLike | None
            The start date of the report.
        end_date: DateLike | None
            The end date of the report.
        **kwargs: Any
            Additional keyword arguments.

        Returns
        -------
        Task[WeightsReportAccessor]
            The task for the weights report.
        """


@docstrings_from_sync
class AsyncWeightsReportApi(
    AsyncReportApi[
        [DateLike | None, DateLike | None],
        AsyncWeightsReportAccessor,
        WeightsReportSettings,
    ]
):

    @abc.abstractmethod
    async def get_maybe_cached(  # noqa: D102
        self,
        start_date: DateLike | None,
        end_date: DateLike | None,
        **kwargs: Any,
    ) -> AsyncWeightsReportAccessor | None: ...

    @abc.abstractmethod
    async def calculate(  # noqa: D102
        self,
        start_date: DateLike | None,
        end_date: DateLike | None,
        **kwargs: Any,
    ) -> AsyncWeightsReportAccessor: ...

    @abc.abstractmethod
    async def calculate_as_task(  # noqa: D102
        self,
        start_date: DateLike | None,
        end_date: DateLike | None,
        **kwargs: Any,
    ) -> AsyncTask[AsyncWeightsReportAccessor]: ...


class WeightsReportAccessorImpl(WeightsReportAccessor):  # noqa: D101

    def __init__(self, accessor: ReportAccessorApi):
        self._accessor = accessor

    @property
    def accessor(self) -> ReportAccessorApi:  # noqa: D102
        return self._accessor


class AsyncWeightsReportAccessorImpl(AsyncWeightsReportAccessor):  # noqa: D101

    def __init__(self, accessor: AsyncReportAccessorApi):
        self._accessor = accessor

    @property
    def accessor(self) -> AsyncReportAccessorApi:  # noqa: D102
        return self._accessor
